from ..utils import *
from ..base import clustering_model

class simple_k_means(clustering_model):
    def __init__(self, x, k, max_iters=100, seed=None):
        super(simple_k_means, self).__init__(x)
        self.n_clusters = k
        if seed != None: np.random.seed(seed)
        outp = np.random.randint(k, size=(x.shape[0],))
        inpt = np.zeros((x.shape[0],))
        means = self._calc_means(outp)
        for j in range(max_iters):
            if (inpt == outp).all(): break
            inpt = outp.copy()
            means_new = self._calc_means(inpt)
            means = self._replace_means(means, means_new)
            outp = simple_k_means._get_groups(x, means)
        self.iters = j
        self._means = means

    @staticmethod
    def _get_groups(x, means):
        outp = [x - means[i, :] for i in range(means.shape[0])]
        outp = [(item ** 2).sum(1) for item in outp]
        return np.array(outp).argmin(0)

    def _replace_means(self, means, means_new):
        for i in range(self.n_clusters):
            if not np.isnan(means_new[i, :]).all():
                means[i, :] = means_new[i, :]
        return means

    def cluster(self, newx):
        return simple_k_means._get_groups(newx, self._means)


class k_means(simple_k_means):
    def __init__(self, x, k, max_iters=100, seed=None, n_start=10):
        if seed != None: np.random.seed(seed)
        clustering_model.__init__(self, x)
        outp = {}
        for i in range(n_start):
            model = simple_k_means(x, k=k, max_iters=max_iters, seed=None)
            outp[model.total_within_group_sum_of_squares] = model
        model = outp[min(outp.keys())]
        self.iters = model.iters
        self.means = model.means
        self.n_clusters = k



def total_within_group_sum_of_squares_for_different_k(x: np.ndarray, k_max: int = 10, *args, **kwargs):
    outp = np.full(k_max + 1, np.nan)
    for k in range(1, k_max + 1):
        model = k_means(x, k, *args, **kwargs)
        outp[k] = model.total_within_group_sum_of_squares
    return outp


def select_k(x: np.ndarray, k_max: int = 10, *args, **kwargs):
    twss = total_within_group_sum_of_squares_for_different_k(x, k_max=k_max, *args, **kwargs)
    dwss = -np.diff(twss)
    dwss[0] = np.nansum(dwss) / np.log(k_max)
    ratio = dwss[:-1] / dwss[1:]
    return ratio.argmax() + 1


import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Type
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from .utils import *


class learning_model(ABC):
    def __init__(self, x: np.ndarray) -> None:
        self.x = x
        (self.n_obs, self.n_feat) = x.shape


class supervised_model(learning_model, ABC):
    def __init__(self, x: np.ndarray, y: np.ndarray) -> None:
        super().__init__(x)
        self.y = y


class unsupervised_model(learning_model, ABC): ...


class prediction_model(supervised_model,ABC):
    @abstractmethod
    def predict(self, target:np.ndarray)->np.ndarray: ...

    @property
    def fitted(self)->np.ndarray: return self.predict(self.x)

    @property
    def residuals(self)->np.ndarray: return self.y - self.fitted

    @property
    def residual_sum_of_squares(self)->float: return np.sum(self.residuals ** 2)

    @property
    def mean_squared_error(self)->float: return np.mean(self.residuals ** 2)

    def out_of_sample_mean_squared_error(self, x, y)->float:
        return np.mean((y - self.predict(x)) ** 2)

    @property
    def root_mean_squared_error(self)->float: return np.sqrt(self.mean_squared_error)

    def out_of_sample_root_mean_squared_error(self, x, y)->float:
        return np.sqrt(self.out_of_sample_mean_squared_error(x, y))

    @property
    def r_squared(self)->float:
        return 1 - self.residuals.var() / self.y.var()

    @property
    def adjusted_r_squared(self)->float:
        return 1 - (1 - self.r_squared) * (self.n_obs - 1) / self.degrees_of_freedom

    @property
    def degrees_of_freedom(self)->int:
        return self.n_obs - self.n_feat

    @property
    def residual_variance(self)->float:
        return self.residuals.var() * (self.n_obs - 1) / self.degrees_of_freedom


class classification_model(supervised_model, ABC):
    @abstractmethod
    def predict_proba(self,target): ...

    def __init__(self,x,y):
        super(classification_model,self).__init__(x,y)
        self._n_classes = y.max()+1

    @property
    def n_classes(self): return self._n_classes

    @n_classes.setter
    def n_classes(self, value:int): self._n_classes = value

    def classify(self,target:np.array): return self.predict_proba(target).argmax(1)

    @property
    def fitted(self): return self.classify(self.x)

    @property
    def accuracy(self): return np.mean(self.y==self.fitted)

    def out_of_sample_accuracy(self,x,y): return np.mean(y==self.classify(x))

    @property
    def misclassification_probability(self): return 1-self.accuracy

    def out_of_sample_misclassification_probability(self,x,y): return 1-self.out_of_sample_accuracy(x,y)

    @property
    def confusion_matrix(self): return one_hot_encode(self.y).T@one_hot_encode(self.fitted)

    def out_of_sample_confusion_matrix(self,x,y): return one_hot_encode(y).T@one_hot_encode(self.classify(x))


class clustering_model(unsupervised_model, ABC):
    n_clusters: int
    means: int

    @abstractmethod
    def cluster(self, target:np.ndarray) -> np.ndarray: ...

    @property
    def n_clusters(self) -> int: return self._n_clusters

    @n_clusters.setter
    def n_clusters(self, x:int) -> None: self._n_clusters = x

    @property
    def means(self) -> np.ndarray: return self._means

    @means.setter
    def means(self, x: np.ndarray) -> None: self._means = x

    @property
    def groups(self) -> np.ndarray: return self.cluster(self.x)

    def _calc_means(self, groups: np.ndarray) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            outp = [self.x[groups == i, :].mean(0) for i in range(self.n_clusters)]
        return np.array(outp)

    @property
    def within_group_sum_of_squares(self) -> np.ndarray:
        outp = [self.x[self.groups == i, :] - self.means[i, :] for i in range(self.n_clusters)]
        outp = [(item ** 2).sum() for item in outp]
        return np.array(outp)

    @property
    def total_within_group_sum_of_squares(self) -> float:
        return self.within_group_sum_of_squares.sum()


class distribution_model(unsupervised_model, ABC):
    def pdf(self, target: np.ndarray) -> np.ndarray: ...

    def cdf(self, target: np.ndarray) -> np.ndarray: ...


class dimension_reduction_model(unsupervised_model, ABC):
    def reduce(self, target: np.ndarray) -> np.ndarray: ...

    def out_of_sample_mean_squared_error(self, target: np.ndarray) -> np.ndarray:
        Z = self.reduce(target)
        M = np.eye(target.shape[0])-Z@np.linalg.inv(Z.T@Z)@Z.T
        return np.trace(target.T@M@target)/target.size

    @property
    def mean_squared_error(self) -> np.ndarray:
        return self.out_of_sample_mean_squared_error(self.x)

    def out_of_sample_root_mean_squared_error(self, target: np.ndarray) -> np.ndarray:
        return np.sqrt(self.out_of_sample_mean_squared_error(target))

    @property
    def root_mean_squared_error(self) -> np.ndarray:
        return self.out_of_sample_root_mean_squared_error(self.x)

class supervised_dimension_reduction_model(supervised_model, ABC):
    def reduce_X(self, x_new: np.ndarray) -> np.ndarray:
        ...

    def reduce_Y(self, y_new: np.ndarray) -> np.ndarray:
        ...

    def reduce(self, x_new: Optional[np.ndarray] = None, y_new: Optional[np.ndarray] = None):
        if x_new is None and y_new is None:
            raise ValueError("Provide at least one of x_new or y_new.")
        if x_new is not None and y_new is not None:
            return self.reduce_X(x_new), self.reduce_Y(y_new)
        return self.reduce_X(x_new) if x_new is not None else self.reduce_Y(y_new)



class likelihood_type(Protocol):
    @property
    def log_likelihood(self) -> float: ...

    @property
    def null_likelihood(self) -> float: ...


class likelihood_model(ABC):
    n_feat: int
    n_obs: int

    @property
    @abstractmethod
    def log_likelihood(self) -> float: ...

    @property
    @abstractmethod
    def null_likelihood(self) -> float: ...

    @property
    def aic(self) -> float: return 2 * self.n_feat - 2 * self.log_likelihood

    @property
    def bic(self) -> float: return np.log(self.n_obs) * self.n_feat - 2 * self.log_likelihood

    @property
    def deviance(self) -> float: return 2 * self.log_likelihood - 2 * self.null_likelihood


class parametric_distribution_model(distribution_model, likelihood_model, ABC):
    params: np.ndarray
    x: np.ndarray

    @abstractmethod
    def pdf(self, target: np.ndarray) -> np.ndarray: ...

    @abstractmethod
    def cdf(self, target: np.ndarray) -> np.ndarray: ...

    @property
    def log_likelihood(self) -> float: return self.out_of_sample_log_likelihood(self.x)

    @abstractmethod
    def out_of_sample_log_likelihood(self, target: np.ndarray): ...

    @property
    def null_likelihood(self) -> float: return self.out_of_sample_null_likelihood(self.x)

    @abstractmethod
    def out_of_sample_null_likelihood(self, target: np.ndarray): ...

    @property
    def deviance(self) -> np.ndarray:
        return 2*self.log_likelihood-2*self.null_likelihood

    def out_of_sample_deviance(self, target: np.ndarray) -> np.ndarray:
        return 2*self.out_of_sample_log_likelihood(target)-2*self.out_of_sample_null_likelihood(target)



class prediction_likelihood_model(ABC):
    y: np.ndarray
    n_obs: int
    n_feat: int

    @abstractmethod
    def evaluate_lnL(self, pred: np.ndarray) -> float: ...

    @property
    @abstractmethod
    def fitted(self) -> np.ndarray: ...

    @property
    def log_likelihood(self) -> float: return self.evaluate_lnL(self.fitted)

    @property
    def null_likelihood(self) -> float: return self.evaluate_lnL(np.full(self.y.shape, self.y.mean()))

    @property
    def aic(self) -> float: return 2 * self.n_feat - 2 * self.log_likelihood

    @property
    def bic(self) -> float: return np.log(self.n_obs) * self.n_feat - 2 * self.log_likelihood

    @property
    def deviance(self) -> float: return 2 * self.log_likelihood - 2 * self.null_likelihood


class broom_model(Protocol):
    @property
    def tidy(self) -> pd.DataFrame: ...

    @property
    def glance(self) -> pd.DataFrame: ...


class variance_model(ABC):
    _glance: pd.DataFrame

    @abstractmethod
    def vcov_params(self) -> np.ndarray: ...

    @property
    def std_error(self): return np.sqrt(np.diagonal(self.vcov_params))

    @property
    def t_statistic(self): return self.params / self.std_error

    @property
    def p_value(self): return 2 * sp.stats.t.cdf(-np.abs(self.t_statistic), df=self.n_obs - self.n_feat)

    def conf_int(self, level=0.95):
        spread = -self.std_error * sp.stats.t.ppf((1 - level) / 2, df=self.degrees_of_freedom)
        return vstack(self.params - spread, self.params + spread)

    @property
    def tidy(self): return self.tidyci(ci=False)

    def tidyci(self, level=0.95, ci=True):
        n = self.n_feat
        df = [np.arange(n), self.params[:n], self.std_error[:n], self.t_statistic[:n], self.p_value[:n]]
        cols = ['variable', 'estimate', 'std.error', 't.statistic', 'p.value']
        if ci:
            df += [self.conf_int(level)[:, :n]]
            cols += ['ci.lower', 'ci.upper']
        df = pd.DataFrame(np.vstack(df).T, columns=cols)
        return df

    @property
    def glance(self) -> pd.DataFrame: return self._glance

    @glance.setter
    def glance(self, x: pd.DataFrame) -> None: self._glance = x


class FormulaNotationMixin:
    """
    Mixin for the SupervisedModel wrapper that enables:
      - SupervisedModel.from_formula("y ~ x1 + x2 + ...", data, model_type, ...)
      - Predicting / classifying directly from DataFrames that match the formula.
    Usage:
      class LM(FormulaNotationMixin, SupervisedModel):
          pass
      m = LM.from_formula("y ~ x1 + x2", df, model_type=least_squares_regressor)
    """
    # ---------- Constructors ----------
    @classmethod
    def from_formula(cls,
                     formula: str,
                     data: pd.DataFrame,
                     model_type: Type[supervised_model],
                     *args, **kwargs) -> "SupervisedModel":
        # parse returns (x_vars, y_var, conditionals, processed_df)
        x_vars, y_var, conditionals, processed = parse(formula, data)

        # construct the standard wrapper with parsed args
        obj = cls(x_vars=x_vars,
                  y_var=y_var,
                  data=processed,
                  model_type=model_type,
                  *args, **kwargs)

        # keep metadata for later use
        obj.formula: str = formula
        obj.conditionals: list[str] = conditionals
        return obj

    # ---------- Helpers ----------
    def _process_new_data(self, new_data: pd.DataFrame) -> pd.DataFrame:
        """
        Re-apply the stored formula to a new DataFrame to rebuild derived columns,
        interactions, and the intercept column, then return the processed subset.
        """
        if not hasattr(self, "formula"):
            raise AttributeError("This instance has no stored formula. "
                                 "Construct it with .from_formula().")
        x_vars, y_var, conditionals, processed = parse(self.formula, new_data.copy())
        # Ensure we use the same x_vars order learned at fit time:
        missing = [c for c in self.x_vars if c not in processed.columns]
        if missing:
            raise ValueError(f"New data is missing required columns after parsing: {missing}")
        return processed[self.x_vars]

    # ---------- Convenience Methods (DF in → ndarray out) ----------
    def design_matrix(self, new_data: pd.DataFrame) -> np.ndarray:
        """Return the X matrix for new_data according to the stored formula."""
        return self._process_new_data(new_data).values

    def predict_from_df(self, new_data: pd.DataFrame) -> np.ndarray:
        """Predict using underlying regression-like models (that implement .predict)."""
        if not hasattr(self.model, "predict"):
            raise AttributeError("Underlying model does not support .predict().")
        X = self.design_matrix(new_data)
        return self.model.predict(X)

    def classify_from_df(self, new_data: pd.DataFrame) -> np.ndarray:
        """Classify using underlying classifiers (that implement .classify or .predict_proba)."""
        if hasattr(self.model, "classify"):
            X = self.design_matrix(new_data)
            return self.model.classify(X)
        if hasattr(self.model, "predict_proba"):
            X = self.design_matrix(new_data)
            return self.model.predict_proba(X).argmax(1)
        raise AttributeError("Underlying model does not support classification methods.")


class SupervisedModel(FormulaNotationMixin, ABC):
    def __init__(self, x_vars: list[str], y_var: str, data: pd.DataFrame, model_type: Type[supervised_model], *args, **kwargs) -> None:
        self.model_type: Type[supervised_model] = model_type
        self.model: supervised_model = model_type(data[x_vars], data[y_var], *args, **kwargs)
        self.x_vars: list[str] = x_vars
        self.y_var: str = y_var
        self.data: pd.DataFrame = data

    def tidyci(self, level: float = 0.95, ci: bool = True) -> pd.DataFrame:
        outp = self.model.tidyci(level, ci)
        outp['variable'] = pd.Series(self.x_vars)
        return outp

    @property
    def tidy(self) -> pd.DataFrame: return self.tidyci(ci=False)

    @property
    def glance(self) -> pd.DataFrame: return self.model.glance


class PredictionModel(SupervisedModel, ABC):
    def predict(self, target: np.ndarray) -> np.ndarray: return self.model.predict(target)


class ClassificationModel(SupervisedModel, ABC):
    def classify(self, target: np.ndarray) -> np.ndarray: return self.model.classify(target)


import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Type
from abc import ABC, abstractmethod

from ..base import prediction_model, SupervisedModel
from ..utils import *
from .glm import least_squares_regressor
from ..Distribution import multinomial, likelihood_ratio_test

class recursive_partitioning_regressor(prediction_model):
    def __init__(self, x, y, sign_level=0.95, max_level=None, random_x=False, level=''):
        self._col_indx = np.random.permutation(x.shape[1])[
                         :int(np.round(np.sqrt(x.shape[1])))] if random_x else np.arange(x.shape[1])
        x = x[:, self._col_indx]
        super(recursive_partitioning_regressor, self).__init__(x, y)
        self.max_level, self._level = max_level, level
        self._split_variable, self._split_value, self._p_value, self._left, self._right = np.nan, np.nan, np.nan, None, None
        self._terminal_prediction = y.mean()
        if max_level != None and len(level) + 1 == max_level: return
        outp = np.array([self.__get_RSS(self.x[:, i]) for i in range(self.n_feat)])
        if np.isnan(outp[:, 0]).all(): return
        self._split_variable = np.nanargmin(outp[:, 0])
        self._split_value = outp[self._split_variable, 1]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._p_value = self._get_p_value()
        if max_level == None and self._p_value > (1 - sign_level) / 2 ** (2 * len(self._level) + 2): return
        left_indx = x[:, self._split_variable] == 0 if np.isnan(self._split_value) else x[:,self._split_variable] <= self._split_value
        self._left = recursive_partitioning_regressor(x[left_indx, :], y[left_indx], sign_level=sign_level, max_level=max_level, level=level + 'L')
        self._right = recursive_partitioning_regressor(x[~left_indx, :], y[~left_indx], sign_level=sign_level, max_level=max_level, level=level + 'R')

    def __get_RSS(self, var, binary=False):
        if var.var() == 0: return np.nan, np.nan
        if binary or np.logical_or(var == 0, var == 1).all():
            retp = [self.y[var == 0], self.y[var == 1]]
            if any([item.shape[0] == 0 for item in retp]): return np.nan, np.nan
            outp = [item.mean() for item in retp]
            retp = [((retp[i] - outp[i]) ** 2).sum() for i in range(2)]
            return sum(retp), np.nan
        splits = np.linspace(var.min(), var.max(), 10)
        RSS = np.array([self.__get_RSS(var > split, True)[0] for split in splits])
        if np.isnan(RSS).all(): return np.nan, np.nan
        indx = np.nanargmin(RSS)
        return RSS[indx], splits[indx]

    def _get_p_value(self):
        xvar = self.x[:, self._split_variable] == 1 if np.isnan(self._split_value) else self.x[:,
                                                                                        self._split_variable] > self._split_value
        xvar = hstack(np.ones(self.n_obs), xvar.astype(int))
        try:
            model = least_squares_regressor(xvar, self.y)
        except:
            return 1
        return model.p_value[1]

    def __str__(self):
        if self._left == None: return f'level:{self._level}; n.obs:{self.n_obs}; residual.ss:{self.residual_sum_of_squares}; variable:{np.nan}; split:{np.nan}; prediction:{self._terminal_prediction}; mean.squared.error:{self.mean_squared_error}; p.value:{self._p_value}\n'
        return f'level:{self._level}; n.obs:{self.n_obs}; residual.ss:{self.residual_sum_of_squares}; variable:{self._col_indx[self._split_variable]}; split:{self._split_value}; prediction:{np.nan}; mean.squared.error:{self.mean_squared_error}; p.value:{self._p_value}\n' + str(
            self._left) + str(self._right)

    def predict(self, newx, fitted=False):
        outp = np.full(shape=newx.shape[0], fill_value=self._terminal_prediction)
        if self._left == None: return outp
        if not fitted: newx = newx[:, self._col_indx]
        left_indx = newx[:, self._split_variable] == 0 if np.isnan(self._split_value) else newx[:,
                                                                                           self._split_variable] <= self._split_value
        outp[left_indx], outp[~left_indx] = self._left.predict(newx[left_indx, :]), self._right.predict(
            newx[~left_indx, :])
        return outp

    @property
    def fitted(self):
        return self.predict(self.x, fitted=True)

    @property
    def tidy(self):
        outp = str(self)
        for i, line in enumerate(outp.split('\n')[:-1]):
            line = line.split(';')
            line = [item.strip().split(':') for item in line]
            mydict = {key: value for key, value in line}
            level = mydict.pop('level')
            newdf = pd.DataFrame(mydict, index=[level])
            if i == 0:
                df = newdf
            else:
                df = pd.concat([df, newdf], axis=0)
        df.index.name = 'level'
        for column in df.columns:
            if column in ['n.obs']:
                df[column] = df[column].astype(int)
            elif column in ['variable']:
                df[column] = df[column].astype(str)
            else:
                df[column] = df[column].astype(float)
        return df

    @property
    def glance(self):
        return pd.DataFrame(self._glance_dict, index=[''])

    @property
    def _glance_dict(self):
        return {'resid.sum.squares': self.residual_sum_of_squares,
                'mean.squared.error': self.mean_squared_error,
                'root.mean.squared.error': self.root_mean_squared_error,
                'r.squared': self.r_squared,
                'adjusted.r.squared': self.adjusted_r_squared,
                'self.df': self.n_feat,
                'resid.df': self.degrees_of_freedom,
                'resid.var': self.residual_variance}



from itertools import product
import numpy as np
import cvxopt
from ..utils import *
from ..base import *
from .glm import linear_model, least_squares_regressor
from typing import Optional
from functools import partial

cvxopt.solvers.options['show_progress'] = False

class l1_regularization_regressor(linear_model):
    def __init__(self, x, y, thresh: float, *args, **kwargs):
        super().__init__(x, y, thresh=thresh, *args, **kwargs)
        self.threshold = thresh

    def _fit(self, x, y, thresh: float, *args, **kwargs):
        if np.all(x[:, 0] == 1):
            dx = x[:, 1:] - x[:, 1:].mean(0)
            dy = y - y.mean(0)
            outp = l1_regularization_regressor.solve_lasso(dx, dy, thresh)
            intc = y.mean(0) - x[:, 1:].mean(0) @ outp.reshape(-1, 1)
            return np.concatenate([intc, outp])
        elif np.all(x[:, -1] == 1):
            dx = x[:, :-1] - x[:, :-1].mean(0)
            dy = y - y.mean(0)
            outp = l1_regularization_regressor.solve_lasso(dx, dy, thresh)
            intc = y.mean(0) - x[:, -1:].mean(0) @ outp.reshape(-1, 1)
            return np.concatenate([outp, intc])
        else:
            return l1_regularization_regressor.solve_lasso(x, y, thresh)

    @staticmethod
    def solve_lasso(x, y, thresh):
        r = x.shape[1]
        P = np.kron(np.array([[1, -1], [-1, 1]]), x.T @ x)
        q = -np.kron(np.array([[1], [-1]]), x.T @ y.reshape(-1, 1))
        G = np.vstack([-np.eye(2 * r), np.ones((1, 2 * r))])
        h = np.vstack([np.zeros((2 * r, 1)), np.array([[thresh]])])
        b = np.array(cvxopt.solvers.qp(*[cvxopt.matrix(i) for i in [P, q, G, h]])['x'])
        return b[:r, 0] - b[r:, 0]


class l1_cross_validation_regressor(l1_regularization_regressor):
    def __init__(self, x, y, max_thresh: Optional[int] = None, folds: int = 5, seed=None, *args, **kwargs):
        default_state = cvxopt.solvers.options.get('show_progress', True)
        cvxopt.solvers.options['show_progress'] = False
        if max_thresh == None: max_thresh = np.abs(linear_model(x, y).params[1:]).sum()
        cv = k_fold_cross_validation(x, y, folds=folds, seed=seed)
        lam_values = np.linspace(0, 1, 100)
        mses = np.zeros(lam_values.shape[0])
        for i,lam in enumerate(lam_values):
            mses[i] = sum([l1_regularization_regressor(x_train, y_train, thresh=lam * max_thresh).out_of_sample_mean_squared_error(x_test, y_test) for x_train,x_test,y_train,y_test in cv])/folds
        i = mses.argmin()
        cvxopt.solvers.options['show_progress'] = default_state
        super().__init__(x, y, thresh=lam_values[i] * max_thresh, *args, **kwargs)
        self.statistic = mses[i]
        self.lambda_value = lam_values[i]
        self.max_threshold = max_thresh


class l1_bootstrap_regressor(l1_cross_validation_regressor,variance_model):
    def __init__(self,x,y,*args,bootstraps:int=1000,**kwargs):
        super().__init__(x,y,*args,**kwargs)
        self.n_boot = bootstraps
        model = lambda x,y: l1_regularization_regressor(x,y,thresh=self.threshold)
        default_state = cvxopt.solvers.options.get('show_progress',True)
        cvxopt.solvers.options['show_progress'] = False
        self.bootstraps = bootstrap(model,x,y,bootstraps=bootstraps)
        cvxopt.solvers.options['show_progress'] = default_state
        self.bootstrap_params = np.array([item.params for item in self.bootstraps])
        self.glance = pd.DataFrame({'r.squared':self.r_squared,
                                        'adjusted.r.squared':self.adjusted_r_squared,
                                        'self.df':self.n_feat,
                                        'resid.df':self.degrees_of_freedom,
                                        'aic':self.aic,
                                        'bic':self.bic,
                                        'log.likelihood':self.log_likelihood,
                                        'deviance':self.deviance,
                                        'resid.var':self.residual_variance},index=[''])
    @property
    def vcov_params(self):
        x = self.bootstrap_params-self.bootstrap_params.mean(0)
        return x.T@x/self.n_boot

L1BootstrapRegressor = partial(PredictionModel,model_type=l1_bootstrap_regressor)



import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Iterable, Tuple, Type
from abc import ABC, abstractmethod
from dataclasses import dataclass
from .base import supervised_model

def one_hot_encode(x: np.ndarray) -> np.ndarray:
    n = x.shape[0]
    ohe = np.zeros((n, x.max() + 1))
    ohe[np.arange(n), x] = 1
    return ohe


def itemfreq(x: np.ndarray, axis: Optional[int] = None, classes: Optional[int] = None) -> np.ndarray:
    return np.array([(x == i).sum(axis=axis) for i in range(x.max() + 1 if classes==None else classes)]).T

def itemprob(x: np.ndarray, axis: Optional[int] = None, classes: Optional[int] = None) -> np.ndarray:
    return np.array([(x == i).mean(axis=axis) for i in range(x.max() + 1 if classes==None else classes)]).T


def expit(x: np.ndarray) -> np.ndarray:
    # Use a piecewise definition for numerical stability
    out = np.empty_like(x, dtype=np.float64)

    pos_mask = x >= 0
    neg_mask = ~pos_mask

    # For positive x, compute normally
    out[pos_mask] = 1 / (1 + np.exp(-x[pos_mask]))

    # For negative x, rewrite sigmoid to avoid overflow in exp(-x)
    exp_x = np.exp(x[neg_mask])
    out[neg_mask] = exp_x / (1 + exp_x)

    return out

def hstack(*args: Tuple[np.ndarray])->np.ndarray:
    return np.hstack([item if item.ndim==2 else item.reshape(-1, 1) for item in args])


def vstack(*args: np.ndarray)->np.ndarray:
    return np.vstack([item if item.ndim==2 else item.reshape(-1, 1) for item in args])


def bind(*args: np.ndarray)->np.ndarray:
    return np.concatenate(args)


def grid_search(func: Callable[[np.ndarray],np.ndarray], space: np.ndarray, axis: Optional[int] = None, maximize: bool = False):
    return func(space).argmax(axis=axis) if maximize else func(space).argmin(axis=axis)


def gradient_descent(gradient: Callable[[np.ndarray],np.ndarray], init_x: np.ndarray, learning_rate: float = 0.005, threshold: float = 1e-10, max_reps: int = 10000, maximize: bool = False):
    x = init_x.copy()
    for i in range(max_reps):
        gx = gradient(x)
        x += gx * learning_rate if maximize else -gx * learning_rate
        if np.abs(gx).sum() < threshold: return x, True
    return x, False


def newton(gradient: Callable[[np.ndarray],np.ndarray], hessian: Callable[[np.ndarray],np.ndarray], init_x: np.ndarray, max_reps: int = 100, tolerance: float = 1e-14):
    x = init_x.copy()
    for i in range(max_reps):
        hess = hessian(x)
        grad = gradient(x)
        try:
            update = -np.linalg.solve(hess, grad)
        except:
            return (x, i - 1)
        x += update
        if np.abs(update).sum() < tolerance: return (x, i)
    raise Exception('Newton did not converge')


def add_intercept(x_vars: List[np.ndarray], y_var: str, data: pd.DataFrame):
    newdf = data.copy()
    x_vars = ['(intercept)'] + x_vars
    newdf['(intercept)'] = np.ones((data.shape[0],))
    return (x_vars, y_var, newdf)



def test_train_split(x: np.ndarray, y: np.ndarray, test_ratio: float = 0.1, seed: Optional[int] = None):
    n = x.shape[0]
    n_test = int(n * test_ratio)
    n_train = n - n_test
    if seed != None: np.random.seed(seed)
    shuffle = np.random.permutation(n)
    x_test = x[shuffle, :][:n_test, :]
    x_train = x[shuffle, :][n_test:, :]
    y_test = y[shuffle][:n_test]
    y_train = y[shuffle][n_test:]
    return (x_train, x_test, y_train, y_test)


def test_split_pandas(data: pd.DataFrame, seed: Optional[int] = None, test_ratio: float = 0.1):
    n = data.shape[0]
    n_test = int(n * test_ratio)
    if seed != None: np.random.seed(seed)
    shuffle = np.random.permutation(n)
    x_test = data.iloc[shuffle[:n_test], :]
    x_train = data.iloc[shuffle[n_test:], :]
    return x_train, x_test


def k_fold_cross_validation(x: np.ndarray, y: np.ndarray, folds: int = 5, seed: Optional[int]=None):
    n = x.shape[0]
    deck = np.arange(n)
    if seed is not None: np.random.seed(seed)
    np.random.shuffle(deck)
    outp = []
    for i in range(folds):
        test = deck[int(i * n / folds):int((i + 1) * n / folds)]
        train_lower = deck[:int(i * n / folds)]
        train_upper = deck[int((i + 1) * n / folds):]
        train = np.concatenate((train_lower, train_upper))
        outp += [(x[train,:],x[test,:],y[train],y[test])]
    return outp

def bootstrap(model:Type[supervised_model],x:np.ndarray,y:np.ndarray,seed:Optional[int]=None,bootstraps:int=1000):
    outp = []
    if seed is not None: np.random.seed(seed)
    for i in range(bootstraps):
        sample = np.random.randint(x.shape[0],size=(x.shape[0],))
        outp += [model(x[sample],y[sample])]
    return outp

def set_product(*args: Tuple[Iterable]) -> list:
    def bind_item(item, this_iterable):
        return [bind_item(item, item) for item in this_iterable] if isinstance(this_iterable[0][0], Iterable) else [(item,) + item for item in this_iterable]
    return [(x,) for x in args[0]] if len(args) == 1 else [bind_item(x,set_product(*args[1:])) for x in args[0]]

def intercept(x: np.ndarray) -> np.ndarray:
    return hstack(np.ones(x.shape[0]),x)

def C(n: int, r: int) -> np.ndarray:
    if n==r: return np.ones((1,r))
    if r==1: return np.eye(n)
    if n<r: raise Exception('Invalid input n<r')
    top_right = C(n-1,r-1)
    bottom_right = C(n-1,r)
    top = cds.hstack(np.ones(top_right.shape[0]),top_right)
    bottom = cds.hstack(np.zeros(bottom_right.shape[0]),bottom_right)
    return cds.vstack(top,bottom)


import numpy as np
from typing import Type, Dict
from ..base import distribution_model
from ..utils import *


## BOOTSTRAP THE FINITE SAMPLE DISTRIBUTION
def likelihood_ratio_test(null: distribution_model, alternative: distribution_model) -> Dict[str, float]:
    lrt_stat = null.log_likelihood - alternative.log_likelihood
    lrt_stat = np.exp(lrt_stat)
    p_value = sp.stats.chi2.cdf(lrt_stat, df=len(alternative.params) - len(null.params))
    return {'likelihood.ratio.statistic': lrt_stat, 'p.value': p_value}



from ..base import *
from ..utils import *
from .glm import logistic_classifier
from .tree import recursive_partitioning_classifier
from functools import partial

class bagging_logistic_classifier(logistic_classifier):
    def __init__(self, x, y, probability=0.5, seed=None, bootstraps=1000):
        super(bagging_logistic_classifier, self).__init__(x, y, probability=probability)
        self.seed = seed
        self.n_boot = bootstraps
        model = lambda x, y: logistic_classifier(x, y, probability=probability)
        self.bootstraps = bootstrap(model, x, y, seed=seed, bootstraps=bootstraps)
        self.bootstrap_params = np.array([item.model.params for item in self.bootstraps])
        self.params = self.bootstrap_params.mean(0)
        vcov_params = (self.bootstrap_params - self.params).T @ (self.bootstrap_params - self.params) / self.n_boot
        self.model = abstract_logistic_regressor(x, y, self.params, vcov_params)

    def predict_proba(self, newx):
        outp = (np.array([item.classify(newx) for item in self.bootstraps]).mean(0) > 0.5).astype(int)
        return hstack(1 - outp, outp)

# TEST AVERAGING PROBABILITIES VS USING THE CLASSIFICATIONS
class bagging_recursive_partitioning_classifier(recursive_partitioning_classifier):
    def __init__(self, x, y, seed=None, bootstraps=1000, sign_level=0.95, max_level=2, random_x=False, weights = None):
        super(bagging_recursive_partitioning_classifier, self).__init__(x, y, max_level=1)
        self.seed, self.n_boot = seed, bootstraps
        model = lambda x, y: recursive_partitioning_classifier(x, y, sign_level=sign_level, max_level=max_level,
                                                               random_x=random_x, weights=weights)
        self.bootstraps = bootstrap(model, x, y, seed=seed, bootstraps=bootstraps)
        fit = self.fitted
        indx = np.array([np.mean(item.fitted == fit) for item in self.bootstraps]).argmax()
        model = self.bootstraps[indx]
        for key, value in vars(model).items(): setattr(self, key, value)

    def predict_proba(self, target, fitted=False):
        return itemprob(np.array([item.classify(target, fitted=fitted) for item in self.bootstraps]), axis=0)


random_forest_classifier = partial(bagging_recursive_partitioning_classifier, random_x=True)



import numpy as np
import pandas as pd
from typing import Optional, Tuple, Dict, Any, Callable
import re
from itertools import chain, combinations, combinations_with_replacement
from functools import partial

################################################################################
#################################  Constants  ##################################
################################################################################

USE_CARET: bool = False
LOGGING: bool = False
NUMPY_FUNCS: Dict[str, Callable[[np.ndarray], np.ndarray]] = {'around': lambda x: np.around(x),
                                                              'rint': lambda x: np.rint(x),
                                                              'fix': lambda x: np.fix(x),
                                                              'floor': lambda x: np.floor(x),
                                                              'ceil': lambda x: np.ceil(x),
                                                              'trunc': lambda x: np.trunc(x),
                                                              'diff': lambda x: np.diff(x),
                                                              'ediff1d': lambda x: np.ediff1d(x),
                                                              'exp': lambda x: np.exp(x),
                                                              'expm1': lambda x: np.expm1(x),
                                                              'exp2': lambda x: np.exp2(x),
                                                              'log': lambda x: np.log(x),
                                                              'log10': lambda x: np.log10(x),
                                                              'log2': lambda x: np.log2(x),
                                                              'log1p': lambda x: np.log1p(x),
                                                              'i0': lambda x: np.i0(x),
                                                              'sinc': lambda x: np.sinc(x),
                                                              'signbit': lambda x: np.signbit(x),
                                                              'spacing': lambda x: np.spacing(x),
                                                              'reciprocal': lambda x: np.reciprocal(x),
                                                              'positive': lambda x: np.positive(x),
                                                              'negative': lambda x: np.negative(x),
                                                              'angle': lambda x: np.angle(x),
                                                              'real': lambda x: np.real(x),
                                                              'imag': lambda x: np.imag(x),
                                                              'conj': lambda x: np.conj(x),
                                                              'conjugate': lambda x: np.conjugate(x),
                                                              'sqrt': lambda x: np.sqrt(x),
                                                              'cbrt': lambda x: np.cbrt(x),
                                                              'square': lambda x: np.square(x),
                                                              'absolute': lambda x: np.absolute(x),
                                                              'fabs': lambda x: np.fabs(x),
                                                              'sign': lambda x: np.sign(x),
                                                              'nan_to_num': lambda x: np.nan_to_num(x),
                                                              'real_if_close': lambda x: np.real_if_close(x)}

################################################################################
##############################  Helper functions  ##############################
################################################################################

def bind(x: list) -> list:
    return list(chain(*x))

def unique(x: list) -> list:
    outp: list = []
    for item in x:
        if item not in outp:
            outp.append(item)
    return outp

def split_expression(expression: str, delimiter: str = '+') -> list[str]:
    # Preprocess the expression to remove spaces for easier processing
    expression = expression.replace(' ', '')

    parts: list[str] = []  # To store the split parts of the expression
    current_part: list[str] = []  # To build the current part of the expression
    depth: int = 0  # Track the depth of parentheses nesting

    for char in expression:
        if char == delimiter and depth == 0:
            # Join the characters of the current part and add it to the parts list
            parts += [''.join(current_part)]
            current_part = []
        else:
            depth += 1 if char == '(' else -1 if char == ')' else 0
            current_part += [char]

    # Add the last part to the parts list, if it's not empty
    if current_part: parts += [''.join(current_part)]
    return parts

def match_parens(expression: str) -> Optional[str]:
    if expression[0]!='(' or expression[-1]!=')': return None
    current_depth: int = 0
    depth: list[int] = []
    expression = expression[1:-1]
    for char in expression:
        current_depth += 1 if char == '(' else -1 if char == ')' else 0
        depth += [current_depth]
    if any([item<0 for item in depth]): return None
    return expression

def make_pretty_minus(expression: str) -> str:
    outp: str = '+-'.join(split_expression(expression,'-'))
    outp = outp.replace('++', '+')
    if len(outp)!=0 and outp[0] == '+': outp = outp[1:]
    return outp

def bin(x: list) -> dict:
    return {item:sum([1 for newbie in x if newbie==item]) for item in unique(x)}

################################################################################
#############################  Expression Parsing  #############################
################################################################################

def parse(formula: str, data: pd.DataFrame) -> Tuple[list[str], str, list[str], pd.DataFrame]:
    y_var, rhs = formula.split('~')
    if '|' in formula:
        rhs, conditionals = formula.split('|')
    else:
        conditionals = ''
    data = data.copy()
    data['(intercept)'] = 1
    parse_term(y_var, data)
    x_vars_included, x_vars_excluded = parse_expression(rhs, data)
    conditionals_included, conditionals_excluded = parse_expression(conditionals, data)
    x_vars_included = ['(intercept)']+x_vars_included
    x_vars = [item for item in x_vars_included if item not in x_vars_excluded and item != y_var]
    conditionals = [item for item in conditionals_included if item not in conditionals_excluded and item != y_var]
    x_vars = unique(x_vars)
    conditionals = unique(conditionals)
    return x_vars, y_var, conditionals, data[[y_var]+x_vars+conditionals]

def parse_expression(expression: str, data: pd.DataFrame) -> tuple[list[str], list[str]]:
    expression = expression.strip()
    expression = make_pretty_minus(expression)
    expression = expression.replace('np.', '')
    expression = expression.replace('numpy.', '')
    if USE_CARET: expression = expression.replace('^','**')
    if result:=parse_basic(expression, data):
        if LOGGING: print('simple',expression,result[0],result[1])
        return result
    if result:=parse_complex(expression, data):
        if LOGGING: print('complex',expression,result[0],result[1])
        return result
    return None

def parse_basic(expression: str, data: pd.DataFrame) -> Optional[tuple[list[str], list[str]]]:
    if expression == '':
        return [], []
    if match := match_parens(expression):
        return parse_expression(match, data)
    if expression == '1':
        return ['(intercept)'], []
    if parse_term(expression, data):
        return [expression], []
    if expression == '.':
        return data.columns, []
    if match := re.match(r'^\((.*)\)\*\*(\d*)$',expression):
        if match_parens(f'({match.group(1)})'):
            power: int = int(match.group(2))
            expression = match.group(1)
            included: list[str]
            excluded: list[str]
            included, excluded = parse_expression(expression, data)
            if excluded:
                raise ValueError(f'Expression {expression} seems to have excluded terms in a power')
            included += generate_interactions(included, power=power, data=data)
            return included, excluded
    if len(terms := split_expression(expression)) > 1:
        included: list[str] = []
        excluded: list[str] = []
        for term in terms:
            if result := parse_expression(term, data):
                included += result[0]
                excluded += result[1]
        return included, excluded
    if expression[0]=='-':
        if not (result := parse_expression(expression[1:], data)):
            raise ValueError(f'Expression {expression} cannot be inverted')
        return result[1], result[0]
    if included := check_special_power_funcs(expression, data):
        return included, []
    return None

def parse_complex(expression: str, data: pd.DataFrame) -> Optional[Tuple[list[str], list[str]]]:
    included: list[str]
    excluded: list[str] = []

    if result := parse_complex_expression_by_splitting_on_string(expression,data):
        if result['error']:
            raise ValueError(f'Expression {expression} found negations in product')
        distributed: bool = result['distributed']
        included = result['terms']
        if distributed:
            return included, excluded
        included += generate_interactions(included, data=data)
        return included, excluded
    if result := parse_complex_expression_by_splitting_on_string(expression, data, delimiter=':'):
        if result['error']:
            raise ValueError(f'Expression {expression} found negations in interaction')
        included = result['terms']
        expression = ':'.join(included)
        included = [expression]
        if is_interaction(expression):
            return included, excluded
        else:
            raise ValueError(f'Expression {expression} cannot parse an interaction')
    return None

def parse_complex_expression_by_splitting_on_string(expression: str, data: pd.DataFrame, delimiter: str = '*') -> Optional[Dict[str,Any]]:
    mod_expression: str = expression.replace('**','<SPECIAL_DELIMITER>')
    if len(terms := split_expression(mod_expression, delimiter=delimiter))>1:
        new_terms: list = []
        for i,term in enumerate(terms):
            unmod_term = term.replace('<SPECIAL_DELIMITER>','**')
            if result := parse_basic(unmod_term, data):
                included = result[0]
                excluded = result[1]
                if excluded:
                    return {'error':True, 'distributed':False, 'terms':None}
                if len(included)>1:
                    retp = [terms[:i]+[included[j]]+terms[(i+1):] for j in range(len(included))]
                    retp = [parse_expression(delimiter.join(item).replace('<SPECIAL_DELIMITER>', '**'),data) for item in retp]
                    if bind([item[1] for item in retp]):
                        return {'error':True, 'distributed':False, 'terms':None}
                    return {'error': False, 'distributed': True, 'terms': bind([item[0] for item in retp])}
                elif len(included)==1:
                    new_terms += [unmod_term]
                else:
                    ...
        return {'error': False, 'distributed': False, 'terms': new_terms}
    return None

################################################################################
################################  Term parsing  ################################
################################################################################

def parse_term(term: str, data: pd.DataFrame) -> bool:
    term = term.strip()
    if in_numpy(term, data): return True
    if is_interaction(term, data): return True
    if is_as_is(term, data): return True
    return False

def is_interaction(expression: str, data: pd.DataFrame) -> bool:
    if len(terms:=split_expression(expression, ':'))>1:
        outp = data['(intercept)'].copy()
        for term in terms:
            if not parse_term(term, data):
                return False
            outp *= data[term]
        data[expression] = outp
        return True
    return False

def in_numpy(expression: str, data: pd.DataFrame) -> bool:
    if expression in data.columns: return True
    for name, function in NUMPY_FUNCS.items():
        pattern = r'^' + name + '\((.*)\)$'
        if match := re.match(pattern, expression):
            inside: str = match.group(1)
            if inside in data.columns:
                data[expression] = function(data[inside])
                return True
    return False

def is_as_is(expression: str, data: pd.DataFrame) -> bool:
    if expression[0]!='I': return False
    if match := match_parens(expression[1:]):
        for item in data.columns:
            match = match.replace(item, f'data[\'{item}\']')
        for name in NUMPY_FUNCS:
            if not re.search(r'\.'+name+r'\(', match) and re.search(name+r'\(', match):
                match = match.replace(name,'np.'+name)
        data[expression] = eval(match)
        return True
    return False

################################################################################
#########################  Generation of Interactions  #########################
################################################################################

def generate_interactions(x: list[str], data: pd.DataFrame, power: Optional[int] = None) -> list[str]:
    x = unique(x.copy())
    if power==None: power = len(x)
    x = unique(bind([unique([":".join(item) for item in combinations(x, r=i)]) for i in range(2, power + 1)]))
    for i, item in enumerate(x):
        terms = split_expression(item, delimiter=":")
        terms = unique(terms)
        terms.sort()
        x[i] = ":".join(terms)
    x = unique(x)
    if any([not is_interaction(item, data) for item in x]):
        raise ValueError(f'Terms {x} failed to generate interactions')
    return x

################################################################################
##########################  Special power functions  ###########################
################################################################################

def special_power(terms: list[str], data: pd.DataFrame, power: int = 1) -> Optional[list[str]]:
    outp: list[str] = []
    terms += ['1']
    for pairing in combinations_with_replacement(terms, r=power):
        pairing = [item for item in pairing if item!='1']
        match len(pairing):
            case 0:
                continue
            case 1:
                expression = pairing[0]
            case _:
                pairing.sort()
                mydict = bin(pairing)
                if len(mydict)==1:
                    term = list(mydict.keys())[0]
                    power = mydict[term]
                    expression = f'I({f"({term})**{power}" if len(split_expression(term)) > 1 else f"{term}**{power}" if power != 1 else f"{term}"})'
                else:
                    expression = f'I({"*".join([f"(({key})**{value})" if len(split_expression(key)) > 1 else f"({key}**{value})" if value != 1 else f"{key}" for key, value in bin(pairing).items()])})'

        if parse_term(expression, data):
            outp += [expression]
        else:
            raise ValueError(f'Expression: {expression} could not be parsed as a term')
    return outp

special_power_funcs = {'linear': partial(special_power, power=1),
                       'quadratic': partial(special_power, power=2),
                       'cubic': partial(special_power, power=3),
                       'quartic': partial(special_power, power=4),
                       'quintic': partial(special_power, power=5),
                       'sextic': partial(special_power, power=6),
                       'hexic': partial(special_power, power=6),
                       'septic': partial(special_power, power=7),
                       'octic': partial(special_power, power=8),
                       'nonic': partial(special_power, power=9),
                       'decic': partial(special_power, power=10),
                       'duodecic': partial(special_power, power=12),
                       'vigintic': partial(special_power, power=20)}

def check_special_power_funcs(expression: str, data: pd.DataFrame) -> Optional[list[str]]:
    for key,func in special_power_funcs.items():
        if expression.startswith(key):
            if match:=match_parens(expression[len(key):]):
                terms = split_expression(match, delimiter=',')
                result = func(terms=terms, data=data)
                return result
            else:
                return
        else:
            ...
    return




import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Type, Union, Any
from abc import ABC, abstractmethod

from functools import partial

from ..base import prediction_model, PredictionModel, prediction_likelihood_model, variance_model
from ..utils import *

class linear_model(prediction_model, prediction_likelihood_model):
    def __init__(self, x, y, *args, **kwargs):
        super(linear_model, self).__init__(x, y)
        self.params = self._fit(x, y, *args, **kwargs)

    def _fit(self, x, y, *args, **kwargs): return np.linalg.solve(x.T @ x, x.T @ y)

    def predict(self, newdata): return newdata @ self.params

    def evaluate_lnL(self, pred):
        return -self.n_obs / 2 * (np.log(2 * np.pi * (self.y - pred).var()) + 1)

class logistic_regressor(linear_model,variance_model):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.glance = pd.DataFrame(
            {'mcfaddens.r.squared': self.mcfaddens_r_squared,
             'ben.akiva.lerman.r.squared': self.ben_akiva_lerman_r_squared,
             'self.df': self.n_feat,
             'resid.df': self.degrees_of_freedom,
             'aic': self.aic,
             'bic': self.bic,
             'log.likelihood': self.log_likelihood,
             'deviance': self.deviance},
            index = ['']
        )
    def _fit(self,x,y):
        params,self.iters = newton(self.gradient,self.hessian,np.zeros(self.n_feat))
        return params
    @property
    def vcov_params(self) -> np.ndarray:
        H = self.hessian(self.params)
        try:
            return -np.linalg.inv(H)
        except np.linalg.LinAlgError:
            return -np.linalg.pinv(H)

    def evaluate_lnL(self, pred: np.ndarray) -> float:
        eps = 1e-15  # or 1e-12 depending on precision needs
        pred = np.clip(pred, eps, 1 - eps)
        return self.y.T @ np.log(pred) + (1 - self.y).T @ np.log(1 - pred)
    def gradient(self,coefs):return self.x.T@(self.y-expit(self.x@coefs))
    def hessian(self,coefs):
        x = self.x
        if isinstance(x, (pd.DataFrame, pd.Series)): x = x.values
        Fx = expit(x@coefs)
        inside = np.diagflat(Fx*(1-Fx))
        return -x.T@inside@x
    def predict(self,target):return expit(target@self.params)
    @property
    def mcfaddens_r_squared(self): return 1-self.log_likelihood/self.null_likelihood
    @property
    def ben_akiva_lerman_r_squared(self):
        return (self.y.T@self.fitted+(1-self.y).T@(1-self.fitted))/self.n_obs

    # In glm.py, inside class logistic_regressor

    def marginal_effects(
            self,
            newx: Optional[Union[np.ndarray, pd.DataFrame]] = None,
            average: bool = True,
    ) -> np.ndarray:
        X = self.x if newx is None else newx

        # Accept pandas
        if isinstance(X, (pd.DataFrame, pd.Series)):
            X = X.values

        # Optional convenience: if model has an intercept column of 1s and
        # caller provided X with one fewer column, prepend ones.
        if X.ndim == 2 and X.shape[1] != self.n_feat:
            if self.x.ndim == 2 and np.allclose(self.x[:, 0], 1) and X.shape[1] == self.n_feat - 1:
                X = np.hstack([np.ones((X.shape[0], 1)), X])
            else:
                raise ValueError(f"newx has shape {X.shape}, but model expects {self.n_feat} features.")

        xb = X @ self.params
        Fx = expit(xb)
        slope = Fx * (1.0 - Fx)  # derivative of expit
        effects = slope.reshape(-1, 1) * self.params.reshape(1, -1)
        return effects.mean(0) if average else effects

class least_squares_regressor(linear_model, variance_model):
    def __init__(self, x, y, white: bool = False, hc: int = 3, *args, **kwargs):
        super(least_squares_regressor, self).__init__(x, y, *args, **kwargs)
        self.white = white
        self.hc = hc
        self.glance = pd.DataFrame({'r.squared': self.r_squared,
                                    'adjusted.r.squared': self.adjusted_r_squared,
                                    'self.df': self.n_feat,
                                    'resid.df': self.degrees_of_freedom,
                                    'aic': self.aic,
                                    'bic': self.bic,
                                    'log.likelihood': self.log_likelihood,
                                    'deviance': self.deviance,
                                    'resid.var': self.residual_variance}, index=[''])

    @property
    def vcov_params(self):
        if self.white:
            return self.__white(self.hc)
        return np.linalg.inv(self.x.T @ self.x) * self.residual_variance

    def __white(self, hc):
        e = self.residuals.values if type(self.residuals) == pd.Series else self.residuals
        esq = self.__hc_correction(e ** 2, hc)
        meat = np.diagflat(esq)
        bread = np.linalg.inv(self.x.T @ self.x) @ self.x.T
        return bread @ meat @ bread.T

    def __hc_correction(self, esq, hc):
        mx = 1 - np.diagonal(self.x @ np.linalg.solve(self.x.T @ self.x, self.x.T))
        match hc:
            case 1:
                esq *= self.n_obs / (self.n_obs - self.n_feat)
            case 2:
                esq /= mx
            case 3:
                esq /= mx ** 2
            case 4:
                p = int(np.round((1 - mx).sum()))
                delta = 4 * np.ones((self.n_obs, 1))
                delta = hstack(delta, self.n_obs * (1 - mx) / p)
                delta = delta.min(1)
                esq /= np.power(mx, delta)
            case 5:
                p = int(np.round((1 - mx).sum()))
                delta = max(4, self.n_obs * 0.7 * (1 - mx).max() / p) * np.ones((self.n_obs, 1))
                delta = hstack(delta.reshape(-1, 1), self.n_obs * (1 - mx.reshape(-1, 1)) / p)
                delta = delta.min(1) / 2
                esq /= np.power(mx, delta)
            case _:
                pass
        return esq

    @property
    def _glance_dict(self):
        return



class poisson_regressor(linear_model):
    def _fit(self, x, y):
        params, self.iters = newton(self.gradient, self.hessian, np.zeros(self.n_feat))
        return params

    @property
    def vcov_params(self): return -np.linalg.inv(self.hessian(self.params))

    def evaluate_lnL(self, pred):
        return self.y.T @ np.log(pred) - np.ones((1, self.n_obs)) @ pred + np.ones((1, self.n_obs)) @ np.log(sp.special.factorial(self.y))

    def gradient(self, coefs): return self.x.T @ (self.y - np.exp(self.x @ coefs))

    def hessian(self, coefs):
        Fx = np.exp(self.x @ coefs)
        if type(Fx) == pd.DataFrame or type(Fx) == pd.Series: Fx = Fx.values
        return -self.x.T @ np.diagflat(Fx) @ self.x

    def predict(self, target): return np.exp(target @ self.params)

    @property
    def _glance_dict(self):
        return {'self.df': self.n_feat,
                'resid.df': self.degrees_of_freedom,
                'aic': self.aic,
                'bic': self.bic,
                'log.likelihood': self.log_likelihood,
                'deviance': self.deviance}


LeastSquaresRegressor = partial(PredictionModel,model_type=least_squares_regressor)
LogisticRegressor = partial(PredictionModel,model_type=logistic_regressor)
PoissonRegressor = partial(PredictionModel,model_type=poisson_regressor)

def backward_stepwise(model: Any,
                      criterion: str = "aic",
                      keep_vars: list[str] = None,
                      min_features: int = 1,
                      verbose: bool = False) -> Dict[str, Any]:
    """
    Generic backward stepwise for either supervised_model (x,y)
    or SupervisedModel wrapper (x_vars, y_var, data).
    """
    keep_vars = set(keep_vars or [])

    # unwrap
    if hasattr(model, "x_vars"):   # SupervisedModel
        x_vars, y_var, data = model.x_vars.copy(), model.y_var, model.data
        model_type = model.model_type
        def fit(subset):
            X = data[subset].values
            y = data[y_var].values
            return model_type(X, y)
        feature_names = x_vars
    else:  # supervised_model
        x, y = model.x, model.y
        feature_names = [f"x{i}" for i in range(x.shape[1])]
        def fit(subset_idx):
            return type(model)(x[:, subset_idx], y)

    # scoring
    def score(m):
        val = getattr(m, criterion)
        lower_is_better = criterion.lower() in ["aic","bic","mse","misclassification_probability"]
        return val, lower_is_better

    # initialize
    current = list(range(len(feature_names)))
    fitted = model
    best_score, lower = score(fitted)
    history = [{"step":0,"removed":None,"score":best_score,"features":feature_names.copy()}]
    step = 0

    while len(current) > max(min_features, len(keep_vars)):
        trial = []
        for j in current:
            if feature_names[j] in keep_vars:
                continue
            cand = [idx for idx in current if idx != j]
            try:
                cand_model = fit([feature_names[i] for i in cand] if hasattr(model,"x_vars") else cand)
                sc, _ = score(cand_model)
                trial.append((j, sc, cand_model))
            except Exception as e:
                if verbose: print("skip", feature_names[j], e)
                continue
        if not trial: break

        j_best, sc_best, best_model = min(trial, key=lambda t: t[1]) if lower else max(trial, key=lambda t: t[1])
        improved = (sc_best < best_score if lower else sc_best > best_score)
        if not improved: break

        step += 1
        if verbose:
            print(f"removed {feature_names[j_best]} -> {criterion}={sc_best}")
        current.remove(j_best)
        best_score, fitted = sc_best, best_model
        history.append({"step":step,"removed":feature_names[j_best],"score":best_score,
                        "features":[feature_names[i] for i in current]})

    return {"model":fitted,
            "selected_features":[feature_names[i] for i in current],
            "history":pd.DataFrame(history)}

import numpy as np
import pandas as pd
from typing import Any, Dict, List, Optional

def _is_intercept_col(col: np.ndarray, name: Optional[str]) -> bool:
    if name is not None and name.strip().lower() == '(intercept)':
        return True
    return np.allclose(col, col[0])

def _score_generic(model: Any, criterion: str) -> tuple[float, bool]:
    c = criterion.lower()
    val = getattr(model, c)
    lower_is_better = c in ("aic", "bic", "mse", "misclassification_probability")
    return float(val), lower_is_better

def forward_stepwise(
    model: Any,
    criterion: str = "aic",
    keep_vars: Optional[List[str]] = None,
    max_features: Optional[int] = None,
    prefer_intercept: bool = True,
    verbose: bool = False
) -> Dict[str, Any]:
    """
    Forward stepwise selection starting from intercept-only if present.
    Accepts either:
      - raw supervised_model with model.x, model.y
      - SupervisedModel wrapper with .x_vars, .y_var, .data, .model_type
    Returns: {'model', 'selected_features', 'history'}
    """
    keep_vars = set(keep_vars or [])

    # -------------------------
    # Unwrap the provided model
    # -------------------------
    if hasattr(model, "x_vars"):  # SupervisedModel wrapper
        x_vars: List[str] = model.x_vars.copy()
        y_var: str = model.y_var
        data: pd.DataFrame = model.data
        model_type = model.model_type
        feature_names = x_vars

        def fit(subset_vars: List[str]):
            X = data[subset_vars].values
            y = data[y_var].values
            return model_type(X, y)

        # Find intercept by name or const column
        intercept_name = None
        if "(intercept)" in feature_names:
            intercept_name = "(intercept)"
        elif prefer_intercept:
            for name in feature_names:
                col = data[name].values
                if _is_intercept_col(col, name):
                    intercept_name = name
                    break

        all_pool = feature_names.copy()
        start_set_vars: List[str] = []
        if prefer_intercept and intercept_name:
            start_set_vars = [intercept_name]
            keep_vars.add(intercept_name)

        current_vars = sorted(set(start_set_vars) | keep_vars, key=lambda v: feature_names.index(v) if v in feature_names else 10**9)

        # initial fit on the full model we were given (for baseline score)
        full_model_score, lower_is_better = _score_generic(model.model if hasattr(model, "model") else model, criterion)

    else:  # raw supervised_model
        X = model.x
        y = model.y
        p = X.shape[1]
        feature_names = [f"x{i}" for i in range(p)]

        def fit(subset_idx: List[int]):
            if len(subset_idx) == 0:
                # Can't fit a 0-column model; fallback to best single variable start
                raise ValueError("Empty model: no columns to fit.")
            return type(model)(X[:, subset_idx], y)

        # Detect intercept column (constant)
        intercept_idx = None
        if prefer_intercept:
            for j in range(p):
                if _is_intercept_col(X[:, j], feature_names[j]):
                    intercept_idx = j
                    break

        all_pool_idx = list(range(p))
        current_idx: List[int] = []
        if prefer_intercept and intercept_idx is not None:
            current_idx = [intercept_idx]

        # baseline score from supplied model
        full_model_score, lower_is_better = _score_generic(model, criterion)

    # ----------------------------------------
    # Helper to get candidates and do 1-step add
    # ----------------------------------------
    history = []
    step = 0

    def current_feature_list():
        if hasattr(model, "x_vars"):
            return current_vars.copy()
        return [feature_names[i] for i in current_idx]

    # Initialize with starting (possibly intercept-only or keep_vars)
    try:
        if hasattr(model, "x_vars"):
            init = fit(current_vars if current_vars else [])
        else:
            init = fit(current_idx if current_idx else [])
        best_score, _ = _score_generic(init, criterion)
        best_model = init
    except Exception:
        # If intercept-only (or empty) fails, fall back to greedy best single variable
        trial = []
        if hasattr(model, "x_vars"):
            for v in [v for v in all_pool if v not in keep_vars]:
                try:
                    m = fit([v] + sorted(list(keep_vars)))
                    s, _ = _score_generic(m, criterion)
                    trial.append((v, s, m))
                except Exception:
                    continue
            if not trial:
                raise
            v_best, s_best, m_best = (min if lower_is_better else max)(trial, key=lambda t: t[1])
            current_vars = sorted(set([v_best]) | keep_vars, key=lambda v: feature_names.index(v) if v in feature_names else 10**9)
            best_score, best_model = s_best, m_best
        else:
            pool = [j for j in all_pool_idx if j not in (current_idx)]
            trial = []
            for j in pool:
                try:
                    m = fit(sorted(list(set([j]) | set(current_idx))))
                    s, _ = _score_generic(m, criterion)
                    trial.append((j, s, m))
                except Exception:
                    continue
            if not trial:
                raise
            j_best, s_best, m_best = (min if lower_is_better else max)(trial, key=lambda t: t[1])
            current_idx = sorted(set([j_best]) | set(current_idx))
            best_score, best_model = s_best, m_best

    history.append({"step": step, "added": None, "score": best_score, "features": current_feature_list()})

    # Limit size
    target_max = max_features if max_features is not None else (len(feature_names))

    # ------------------------------------------------
    # Greedy forward: add the single best new variable
    # ------------------------------------------------
    while True:
        if hasattr(model, "x_vars"):
            pool = [v for v in all_pool if v not in current_vars]
            if len(current_vars) >= target_max or len(pool) == 0:
                break

            trials = []
            for v in pool:
                try:
                    cand = current_vars + [v]
                    m = fit(cand)
                    sc, _ = _score_generic(m, criterion)
                    trials.append((v, sc, m))
                except Exception as e:
                    if verbose:
                        print(f"skip add {v}: {e}")
                    continue

            if not trials:
                break
            v_best, s_best, m_best = (min if lower_is_better else max)(trials, key=lambda t: t[1])
            improved = (s_best < best_score) if lower_is_better else (s_best > best_score)
            if not improved:
                break

            current_vars.append(v_best)
            best_score, best_model = s_best, m_best
            step += 1
            if verbose:
                print(f"[+] add {v_best} -> {criterion}={best_score:.6g} (k={len(current_vars)})")
            history.append({"step": step, "added": v_best, "score": best_score, "features": current_feature_list()})

        else:
            pool = [j for j in all_pool_idx if j not in current_idx]
            if len(current_idx) >= target_max or len(pool) == 0:
                break

            trials = []
            for j in pool:
                try:
                    cand = current_idx + [j]
                    m = fit(cand)
                    sc, _ = _score_generic(m, criterion)
                    trials.append((j, sc, m))
                except Exception as e:
                    if verbose:
                        print(f"skip add {feature_names[j]}: {e}")
                    continue

            if not trials:
                break
            j_best, s_best, m_best = (min if lower_is_better else max)(trials, key=lambda t: t[1])
            improved = (s_best < best_score) if lower_is_better else (s_best > best_score)
            if not improved:
                break

            current_idx.append(j_best)
            best_score, best_model = s_best, m_best
            step += 1
            if verbose:
                print(f"[+] add {feature_names[j_best]} -> {criterion}={best_score:.6g} (k={len(current_idx)})")
            history.append({"step": step, "added": feature_names[j_best], "score": best_score, "features": current_feature_list()})

    # Wrap result consistently
    if hasattr(model, "x_vars"):
        selected = current_vars
        final_model = best_model
    else:
        selected = [feature_names[i] for i in current_idx]
        final_model = best_model

    return {
        "model": final_model,
        "selected_features": selected,
        "history": pd.DataFrame(history),
    }

def _metric_value(m: Any, metric: str) -> float:
    """
    Robustly extract metric from either a raw model or a SupervisedModel wrapper.
    Tries attribute on the object, then on .model, then in .glance dataframe.
    """
    metric = metric.lower()
    # direct attribute
    if hasattr(m, metric):
        return float(getattr(m, metric))
    # wrapped model
    if hasattr(m, "model") and hasattr(m.model, metric):
        return float(getattr(m.model, metric))
    # glance fallback
    if hasattr(m, "glance"):
        g = m.glance
        if isinstance(g, pd.DataFrame) and metric in g.columns:
            return float(g.iloc[0][metric])
    raise AttributeError(f"Could not find metric '{metric}' on model.")

def _lower_is_better(metric: str) -> bool:
    metric = metric.lower()
    return metric in ("aic", "bic", "mse", "misclassification_probability")

def _compare_models(m1: Any, m2: Any, metrics: List[str], tol: float = 1e-12) -> Tuple[Any, Dict[str, str]]:
    """
    Return the better of (m1, m2) by a majority vote across metrics.
    Each metric contributes one 'vote'. Ties use the first metric as tie-breaker.
    """
    votes = {"m1": 0, "m2": 0, "ties": 0}
    per_metric = {}
    for met in metrics:
        try:
            v1 = _metric_value(m1, met)
            v2 = _metric_value(m2, met)
        except Exception:
            # if metric not available on either, skip
            continue
        better_is_lower = _lower_is_better(met)
        diff = v1 - v2
        if abs(diff) <= tol:
            votes["ties"] += 1
            per_metric[met] = "tie"
        else:
            if (diff < 0) == better_is_lower:  # m1 better
                votes["m1"] += 1
                per_metric[met] = "m1"
            else:
                votes["m2"] += 1
                per_metric[met] = "m2"

    if votes["m1"] > votes["m2"]:
        winner = m1
    elif votes["m2"] > votes["m1"]:
        winner = m2
    else:
        # tie across votes: fall back to first metric in the list (if available)
        first = metrics[0]
        try:
            v1 = _metric_value(m1, first)
            v2 = _metric_value(m2, first)
            better_is_lower = _lower_is_better(first)
            if (v1 - v2 < 0) == better_is_lower:
                winner = m1
            else:
                winner = m2
        except Exception:
            # if even that fails, default to m1
            winner = m1

    return winner, per_metric

def stepwise(
    model: Any,
    direction: str = "both",
    criterion: str = "aic",
    keep_vars: List[str] | None = None,
    min_features: int = 1,
    max_features: int | None = None,
    prefer_intercept: bool = True,
    verbose: bool = False,
) -> Dict[str, Any]:
    """
    Unified stepwise selector.
    - direction='forwards' → forward_stepwise
    - direction='backwards' → backward_stepwise
    - direction='both' → run both and return the better result by a vote across {criterion, aic, bic}
      (deduplicated if criterion is 'aic' or 'bic').

    Returns:
      {
        'model': best_model,
        'selected_features': list[str],
        'history': pd.DataFrame,           # from the chosen direction
        'direction_chosen': 'forwards'|'backwards',
        'comparison' : {'metric': 'winner', ...}  # only when direction='both'
      }
    """
    direction = direction.lower()
    keep_vars = keep_vars or []

    if direction == "forwards":
        fwd = forward_stepwise(
            model,
            criterion=criterion,
            keep_vars=keep_vars,
            max_features=max_features,
            prefer_intercept=prefer_intercept,
            verbose=verbose,
        )
        return {
            "model": fwd["model"],
            "selected_features": fwd["selected_features"],
            "history": fwd["history"],
            "direction_chosen": "forwards",
        }

    if direction == "backwards":
        bwd = backward_stepwise(
            model,
            criterion=criterion,
            keep_vars=keep_vars,
            min_features=min_features,
            verbose=verbose,
        )
        return {
            "model": bwd["model"],
            "selected_features": bwd["selected_features"],
            "history": bwd["history"],
            "direction_chosen": "backwards",
        }

    if direction != "both":
        raise ValueError("direction must be one of {'forwards','backwards','both'}")

    # run both
    fwd = forward_stepwise(
        model,
        criterion=criterion,
        keep_vars=keep_vars,
        max_features=max_features,
        prefer_intercept=prefer_intercept,
        verbose=verbose,
    )
    bwd = backward_stepwise(
        model,
        criterion=criterion,
        keep_vars=keep_vars,
        min_features=min_features,
        verbose=verbose,
    )

    m_fwd = fwd["model"]
    m_bwd = bwd["model"]

    # build metric set: {criterion, aic, bic} (deduped, criterion first)
    metrics: List[str] = []
    for met in [criterion.lower(), "aic", "bic"]:
        if met not in metrics:
            metrics.append(met)

    winner_model, per_metric = _compare_models(m_fwd, m_bwd, metrics)

    if winner_model is m_fwd:
        chosen = {
            "model": fwd["model"],
            "selected_features": fwd["selected_features"],
            "history": fwd["history"],
            "direction_chosen": "forwards",
            "comparison": per_metric,
        }
    else:
        chosen = {
            "model": bwd["model"],
            "selected_features": bwd["selected_features"],
            "history": bwd["history"],
            "direction_chosen": "backwards",
            "comparison": per_metric,
        }
    return chosen



import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Type
from abc import ABC, abstractmethod

from ..base import prediction_model, SupervisedModel
from ..utils import *

class k_nearest_neighbors_regressor(prediction_model):
    def __init__(self, x: np.array, y: np.array, k: int = 1) -> None:
        super(k_nearest_neighbors_regressor, self).__init__(x, y)
        self.k = k
        self.norms_train = (x ** 2).sum(1).reshape(-1, 1)

    def neighbors(self, target: np.array, k: int) -> np.array:
        norms_test = (target ** 2).sum(1).reshape(-1, 1)
        distance_matrix = self.norms_train - 2 * self.x @ target.T + norms_test.T
        nearest_neighbors = distance_matrix.argsort(0)[:k, :].T
        return nearest_neighbors

    def predict(self, target: np.array) -> np.array:
        nearest_neighbors = self.neighbors(target, self.k)
        return self.y[nearest_neighbors].mean(1)


class k_nearest_neighbors_cross_validation_regressor(k_nearest_neighbors_regressor):
    def __init__(self, x: np.array, y: np.array, k_max: int = 25, folds: int = 5, seed: Optional[int] = None) -> None:
        n = x.shape[0]
        outp = []
        deck = np.arange(n)
        if seed is not None: np.random.seed(seed)
        np.random.shuffle(deck)
        for i in range(folds):
            test = deck[int(i * n / folds):int((i + 1) * n / folds)]
            train_lower = deck[:int(i * n / folds)]
            train_upper = deck[int((i + 1) * n / folds):]
            train = np.concatenate((train_lower, train_upper))
            nearest_neighbors = k_nearest_neighbors_regressor(x[train], y[train], k_max + 1).neighbors(x[test],
                                                                                                       k_max + 1)
            outp += [
                [((y[test] - y[train][nearest_neighbors[:, :k]].mean(1)) ** 2).mean() for k in range(1, k_max + 1)]]
        mspe = np.array(outp).mean(0)
        k = mspe.argmin() + 1
        super(k_nearest_neighbors_cross_validation_regressor, self).__init__(x, y, k)




from ..Prediction import k_nearest_neighbors_regressor
from ..base import *
from ..utils import *


class k_nearest_neighbors_classifier(classification_model, k_nearest_neighbors_regressor):
    def __init__(self, x: np.array, y: np.array, k: int = 1) -> None:
        super(k_nearest_neighbors_classifier, self).__init__(x, y)
        self.k = k
        self.norms_train = (x ** 2).sum(1).reshape(-1, 1)

    def predict_proba(self, target: np.array) -> np.array:
        nearest_neighbors = self.neighbors(target, self.k)
        pred_vals = self.y[nearest_neighbors]
        return itemprob(pred_vals, 1)


class k_nearest_neighbors_cross_validation_classifier(k_nearest_neighbors_classifier):
    def __init__(self, x: np.array, y: np.array, k_max: int = 25, folds: int = 5, seed: Optional[int] = None) -> None:
        n = x.shape[0]
        outp = []
        deck = np.arange(n)
        if seed is not None: np.random.seed(seed)
        np.random.shuffle(deck)
        for i in range(folds):
            test = deck[int(i * n / folds):int((i + 1) * n / folds)]
            train_lower = deck[:int(i * n / folds)]
            train_upper = deck[int((i + 1) * n / folds):]
            train = np.concatenate((train_lower, train_upper))
            nearest_neighbors = k_nearest_neighbors_classifier(x[train], y[train], k_max + 1).neighbors(x[test],
                                                                                                        k_max + 1)
            outp += [[(y[test] == y[train][nearest_neighbors[:, :k]].mean(1)).mean() for k in range(1, k_max + 1)]]
        acc = np.array(outp).mean(0)
        k = acc.argmax() + 1
        super(k_nearest_neighbors_cross_validation_classifier, self).__init__(x, y, k)



import numpy as np
import scipy as sp
import pandas as pd
import warnings
from typing import Optional, Protocol, Callable, List, Dict, Type
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


from ..base import parametric_distribution_model
from ..utils import *

@dataclass(eq=False, order=False)
class two_sample(parametric_distribution_model):
    x: np.ndarray
    y: np.ndarray
    model_type: Type[parametric_distribution_model]
    w_x: Optional[np.ndarray] = None
    w_y: Optional[np.ndarray] = None
    x_model: parametric_distribution_model = field(init=False)
    y_model: parametric_distribution_model = field(init=False)
    params: np.ndarray = field(init=False)

    def __post_init__(self) -> None:
        self.x_model = self.model_type(self.x, self.w_x) if self.w_x is not None else self.model_type(self.x)
        self.y_model = self.model_type(self.y, self.w_y) if self.w_y is not None else self.model_type(self.x)
        self.params = bind(self.x_model.params, self.y_model.params)

    def pdf(self, x: np.ndarray) -> np.ndarray: raise NotImplementedError()

    def cdf(self, x: np.ndarray) -> np.ndarray: raise NotImplementedError()

    @property
    def log_likelihood(self) -> float: return self.x_model.log_likelihood + self.y_model.log_likelihood

    @property
    def null_likelihood(self) -> float: return self.x_model.null_likelihood + self.y_model.null_likelihood

class multinomial(parametric_distribution_model):
    def __init__(self, x: np.ndarray, w_x: Optional[np.ndarray] = None, classes: Optional[int] = None) -> None:
        super().__init__(x.reshape(-1, 1))
        self.n_classes = x.max() + 1 if classes == None else classes
        self.bins = self.n_obs*itemfreq(x*w_x) if w_x is not None else itemfreq(x, classes=self.n_classes)
        self.params = self.bins / self.n_obs
        self.w_x = w_x

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return sp.stats.multinomial.pmf(x, n=1, p=self.params)

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return sp.stats.multinomial.cdf(x, n=1, p=self.params)

    def out_of_sample_log_likelihood(self, target: np.ndarray) -> np.ndarray:
        return self.likelihood_helper(target, self.params)

    def out_of_sample_null_likelihood(self, target: np.ndarray) -> np.ndarray:
        return self.likelihood_helper(target, np.full(self.n_classes, self.n_classes/self.n_obs))

    def likelihood_helper(self,target : np.ndarray, probs: np.ndarray) -> np.ndarray:
        bins = itemprob(target, classes=self.n_classes)
        return sp.special.gammaln(target.sum() + 1) - sp.special.gammaln(bins + 1).sum() - np.log(bins / probs).sum()

class normal(parametric_distribution_model):
    def __init__(self, x: np.ndarray, w_x: Optional[np.ndarray] = None) -> None:
        super().__init__(x.reshape(-1, 1))
        if w_x is not None:
            mean = np.average(self.x, weights=w_x)
            std = np.sqrt(np.average((self.x - self.mean) ** 2, weights=w_x))
        else:
            mean = self.x.mean()
            std = self.x.std(ddof=0)
        self.params = np.array([mean, std])
        self.w_x = w_x

    def pdf(self, target: np.ndarray) -> np.ndarray:
        return sp.stats.norm.pdf(target, loc=self.params[0], scale=self.params[1])

    def cdf(self, target: np.ndarray) -> np.ndarray:
        return sp.stats.norm.cdf(target, loc=self.params[0], scale=self.params[1])

    def out_of_sample_log_likelihood(self, target: np.ndarray) -> float:
        return self.likelihood_helper(target, loc=self.params[0], scale=self.params[1])

    def out_of_sample_null_likelihood(self, target: np.ndarray) -> float:
        null_mean = target.mean()
        null_std = target.std(ddof=0)
        return self.likelihood_helper(target, null_mean, null_std)

    def likelihood_helper(self, target: np.ndarray, mu: float, sigma: float) -> float:
        sigma = max(sigma, 1e-12)
        return sp.stats.norm(loc=mu, scale=sigma).logpdf(target.flatten()).sum()


class uniform(parametric_distribution_model):
    def __init__(self, x: np.ndarray) -> None:
        super().__init__(x.reshape(-1, 1))
        self.params = np.array([self.x.min(), self.x.max()])
        if self.params[1] <= self.params[0]:
            raise ValueError("Uniform distribution requires upper > lower.")

    def pdf(self, target: np.ndarray) -> np.ndarray:
        return sp.stats.uniform.pdf(target, loc=self.params[0], scale=self.params[1]-self.params[0])

    def cdf(self, target: np.ndarray) -> np.ndarray:
        return sp.stats.uniform.cdf(target, loc=self.params[0], scale=self.params[1]-self.params[0])

    def out_of_sample_log_likelihood(self, target: np.ndarray) -> float:
        return self.likelihood_helper(target, self.params[0], self.params[1])

    def out_of_sample_null_likelihood(self, target: np.ndarray) -> float:
        null_lower = target.min()
        null_upper = target.max()
        return self.likelihood_helper(target, null_lower, null_upper)

    def likelihood_helper(self, target: np.ndarray, lower: float, upper: float) -> float:
        if upper <= lower:
            raise ValueError("Uniform distribution requires upper > lower.")
        target = target.flatten()
        in_bounds = (target >= lower) & (target <= upper)
        return -np.inf if not np.all(in_bounds) else -len(target) * np.log(upper - lower)



from ..base import *
from ..utils import *

def _quad_form_rows(X: np.ndarray, A: np.ndarray) -> np.ndarray:
    # einsum is fast and stable for this pattern
    return np.einsum("ij,jk,ik->i", X, A, X)


class linear_discriminant_analysis(classification_model):
    def __init__(self, x: np.array, y: np.array) -> None:
        super(linear_discriminant_analysis, self).__init__(x, y)
        self.mean_vectors = [x[y == i].mean(0).reshape(-1, 1) for i in range(self.n_classes)]
        self.priors = itemprob(y)
        self.Sigma_within = np.zeros((self.n_feat, self.n_feat))
        for i, mean in enumerate(self.mean_vectors):
            for row in x[y == i]:
                row = row.reshape(-1, 1)
                self.Sigma_within += (row - mean) @ (row - mean).T
        self.overall_mean = x.mean(0).reshape(-1, 1)
        self.Sigma_between = np.zeros((self.n_feat, self.n_feat))
        for i, mean in enumerate(self.mean_vectors):
            n_class = x[y == i, :].shape[0]
            self.Sigma_between += n_class * (mean - self.overall_mean) @ (mean - self.overall_mean).T
        self.eigenvalues, self.eigenvectors = np.linalg.eig(np.linalg.solve(self.Sigma_within, self.Sigma_between))
        self.eigenvalues = np.flip(self.eigenvalues)
        self.eigenvectors = np.fliplr(self.eigenvectors)[:, :self.n_classes - 1]
        self.eigenvectors = self.eigenvectors

    def discriminant(self, target: np.array) -> np.array:
        return target @ self.eigenvectors

    def predict_proba(self, target: np.array) -> np.array:
        outp = np.zeros((target.shape[0], self.n_classes))
        Sw_inv = np.linalg.inv(self.Sigma_within)
        for i, mean_vec in enumerate(self.mean_vectors):
            for j, x in enumerate(target):
                tmp = x.reshape(-1, 1) - mean_vec
                outp[j, i] = -0.5 * tmp.T @ Sw_inv @ tmp + np.log(self.priors[i])
        outp -= outp.max(1).reshape(-1, 1)
        outp = np.exp(outp)
        outp /= outp.sum(1).reshape(-1, 1)
        return outp

class quadratic_discriminant_analysis(classification_model):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        priors: Optional[np.ndarray] = None,
        reg: float = 1e-6,
        sample_weight: Optional[np.ndarray] = None,
    ) -> None:
        super().__init__(x, y)

        # Validate sample weights
        if sample_weight is None:
            w = np.ones(self.n_obs, dtype=float)
        else:
            w = np.asarray(sample_weight, dtype=float)
            if w.ndim != 1 or w.shape[0] != self.n_obs:
                raise ValueError("sample_weight must be shape (n_obs,).")
            if np.sum(w) <= 0:
                raise ValueError("sample_weight must sum to a positive value.")

        # Classes and encoded indices
        classes, y_idx = np.unique(y, return_inverse=True)
        self.classes = classes                    # store labels
        self.n_classes = classes.size             # <-- keep this an INT
        k_classes = self.n_classes

        # Priors
        if priors is None:
            cw = np.bincount(y_idx, weights=w, minlength=k_classes).astype(float)
            priors = cw / cw.sum()
        else:
            priors = np.asarray(priors, dtype=float)
            if priors.shape != (k_classes,):
                raise ValueError(f"priors must be shape ({k_classes},)")
            if np.any(priors < 0) or priors.sum() <= 0:
                raise ValueError("priors must be nonnegative and sum to a positive value.")
            priors = priors / priors.sum()
        self.priors = priors

        # Means, covariances, inverses, log determinants
        self.means = np.zeros((k_classes, self.n_feat))
        self.covs = np.zeros((k_classes, self.n_feat, self.n_feat))
        self.inv_covs = np.zeros_like(self.covs)
        self.log_dets = np.zeros(k_classes)

        for k in range(k_classes):
            mask = (y_idx == k)
            wk = w[mask]
            Xk = x[mask]
            W = wk.sum()
            if W <= 0:
                raise ValueError(f"Class {classes[k]} has zero total weight.")

            mu = np.average(Xk, axis=0, weights=wk)
            self.means[k] = mu

            Xc = Xk - mu
            cov = (Xc.T * wk) @ Xc / W  # MLE covariance
            if reg > 0:
                cov = cov + reg * np.eye(self.n_feat)

            sign, logdet = np.linalg.slogdet(cov)
            if sign <= 0:
                cov = cov + max(reg, 1e-8) * np.eye(self.n_feat)
                sign, logdet = np.linalg.slogdet(cov)
                if sign <= 0:
                    raise np.linalg.LinAlgError(
                        f"Covariance for class {classes[k]} is not PD even after regularization."
                    )

            self.covs[k] = cov
            self.inv_covs[k] = np.linalg.inv(cov)
            self.log_dets[k] = logdet

    def decision_function(self, x: np.ndarray) -> np.ndarray:
        self._check_is_fitted()
        x = np.asarray(x, dtype=float)
        if x.ndim != 2:
            raise ValueError("X must be a 2D array of shape (n_samples, n_features).")
        if x.shape[1] != self.n_feat:
            raise ValueError(f"X has {x.shape[1]} features; expected {self.n_feat}")

        k_classes = self.n_classes  # int
        scores = np.zeros((x.shape[0], k_classes))
        for k in range(k_classes):
            mu = self.means[k]
            inv_cov = self.inv_covs[k]
            quad = _quad_form_rows(x - mu, inv_cov)  # needs to exist in utils
            scores[:, k] = np.log(self.priors[k]) - 0.5 * self.log_dets[k] - 0.5 * quad
        return scores

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        scores = self.decision_function(x)
        m = scores.max(axis=1, keepdims=True)
        P = np.exp(scores - m)
        P /= P.sum(axis=1, keepdims=True)
        return P

    def _check_is_fitted(self) -> None:
        attrs = ("means", "covs", "inv_covs", "log_dets", "priors")
        missing = [a for a in attrs if not hasattr(self, a)]
        if missing:
            raise AttributeError(
                f"{self.__class__.__name__} is not fitted. Missing: {', '.join(missing)}"
            )



from ..utils import *
from ..base import dimension_reduction_model, supervised_dimension_reduction_model


class principal_components_analysis(dimension_reduction_model):
    def __init__(self, x: np.ndarray, k: int, *, center: bool = True, scale: bool = False) -> None:
        super().__init__(x)

        if not (1 <= k <= min(self.n_obs, self.n_feat)):
            raise ValueError("k must be between 1 and min(n_obs, n_feat)")

        # training stats
        self.mean = x.mean(0) if center else np.zeros(self.n_feat, dtype=float)
        std = x.std(0, ddof=1)
        if scale:
            # avoid divide-by-zero for constant columns
            self.scale = np.where(std == 0.0, 1.0, std)
        else:
            self.scale = np.ones(self.n_feat, dtype=float)

        x_cs = (x - self.mean) / self.scale

        # SVD: X = U S V^T  -> PCs = columns of V, scores = U S
        U, S, Vt = np.linalg.svd(x_cs, full_matrices=False)
        q = S.shape[0]
        self.singular_values = S
        self.rotation = Vt.T[:, :k]          # (p, k)
        self.components = (U * S)[:, :k]     # (n, k)
        self.k = k

        # variance bookkeeping
        denom = max(self.n_obs - 1, 1)
        ev = (S**2) / denom
        self.explained_variance = ev                       # length q
        total_var = x_cs.var(axis=0, ddof=1).sum()
        self.explained_variance_ratio = ev / total_var if total_var > 0 else np.zeros_like(ev)

    def reduce(self, target: np.ndarray) -> np.ndarray:
        if target.ndim != 2 or target.shape[1] != self.rotation.shape[0]:
            raise ValueError(f"target must be shape (m, {self.rotation.shape[0]})")
        x_cs = (target - self.mean) / self.scale
        return x_cs @ self.rotation


def select_k(eigs: np.ndarray, k_max: int | None = None, include_zero: bool = True, allow_zero: bool = False) -> int:
    lam = np.asarray(eigs, dtype=float)
    lam = lam[~np.isnan(lam)]
    if lam.size < 2:
        return 0 if allow_zero else 1

    lam = np.sort(lam)[::-1]  # λ₁ ≥ λ₂ ≥ ...
    if k_max is None:
        k_max = lam.size
    k_max = max(2, min(k_max, lam.size))  # need at least two to form a ratio

    lam_use = lam[:k_max]                      # [λ₁..λ_{k_max}]
    if include_zero:
        lam0 = lam.sum()/k_max                  # λ₀ := sum(λ₁..λ_{n})/k_max
        lam_aug = np.concatenate(([lam0], lam_use))  # [λ₀, λ₁, …, λ_{k_max}]
        ratios = lam_aug[:-1] / np.maximum(lam_aug[1:], 1e-15)
        # indices 0..k_max-1 correspond to k=0..k_max-1
        k_hat = int(np.argmax(ratios))
        if not allow_zero:
            k_hat = max(1, k_hat)              # enforce ≥1 factor if desired
        return k_hat
    else:
        ratios = lam_use[:-1] / np.maximum(lam_use[1:], 1e-15)  # k=1..k_max-1
        return int(np.argmax(ratios) + 1)


class canonical_correlation_analysis(supervised_dimension_reduction_model):
    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        k: int,
        *,
        center: bool = True,
        scale: bool = False,
        reg: float = 1e-6,
    ) -> None:
        # initialize supervised base
        super().__init__(X, Y)

        n, p = X.shape
        n2, q = Y.shape
        if n != n2:
            raise ValueError("X and Y must have the same number of rows.")
        k_max = min(p, q, max(n - 1, 1))
        if not (1 <= k <= k_max):
            raise ValueError(f"k must be in [1, {k_max}] given shapes X{X.shape}, Y{Y.shape}.")

        # training stats
        self.mean_x = X.mean(0) if center else np.zeros(p, dtype=float)
        self.mean_y = Y.mean(0) if center else np.zeros(q, dtype=float)

        std_x = X.std(0, ddof=1)
        std_y = Y.std(0, ddof=1)
        self.scale_x = np.where(std_x == 0.0, 1.0, std_x) if scale else np.ones(p, dtype=float)
        self.scale_y = np.where(std_y == 0.0, 1.0, std_y) if scale else np.ones(q, dtype=float)

        Xc = (X - self.mean_x) / self.scale_x
        Yc = (Y - self.mean_y) / self.scale_y

        # sample covariance blocks (ddof=1)
        d = max(n - 1, 1)
        Sxx = (Xc.T @ Xc) / d
        Syy = (Yc.T @ Yc) / d
        Sxy = (Xc.T @ Yc) / d

        # --- Whitening matrices Wx = (Sxx+reg I)^(-1/2), Wy = (Syy+reg I)^(-1/2) ---
        # symmetrize then eigendecompose
        Sxx = (Sxx + Sxx.T) * 0.5
        Syy = (Syy + Syy.T) * 0.5

        evalx, evecx = np.linalg.eigh(Sxx)
        evaly, evecy = np.linalg.eigh(Syy)

        # regularize and clamp to avoid zeros/negatives
        tiny = np.finfo(float).tiny
        evalx = np.clip(evalx + reg, tiny, None)
        evaly = np.clip(evaly + reg, tiny, None)

        inv_sqrt_x = evecx @ (np.diag(1.0 / np.sqrt(evalx)) @ evecx.T)
        inv_sqrt_y = evecy @ (np.diag(1.0 / np.sqrt(evaly)) @ evecy.T)
        # re‑symmetrize to remove numerical drift
        inv_sqrt_x = (inv_sqrt_x + inv_sqrt_x.T) * 0.5
        inv_sqrt_y = (inv_sqrt_y + inv_sqrt_y.T) * 0.5

        # whitened cross-covariance
        M = inv_sqrt_x @ Sxy @ inv_sqrt_y

        # --- SVD on whitened matrix ---
        U, Sigma, Vt = np.linalg.svd(M, full_matrices=False)

        # canonical correlations
        self.canonical_correlations = Sigma[:k]

        # canonical directions in original variable space
        A = inv_sqrt_x @ U[:, :k]   # (p, k)
        B = inv_sqrt_y @ Vt.T[:, :k]  # (q, k)

        # Normalize so that a_j^T Sxx a_j = 1 and b_j^T Syy b_j = 1 (unit variance CVs)
        AtS = A.T @ Sxx
        Aj_var = np.sqrt(np.maximum(np.diag(AtS @ A), tiny))
        A = A / Aj_var.reshape(1, -1)
        BtS = B.T @ Syy
        Bj_var = np.sqrt(np.maximum(np.diag(BtS @ B), tiny))
        B = B / Bj_var.reshape(1, -1)

        self.rotation_x = A
        self.rotation_y = B
        self.k = k

        # canonical variates (scores on training data)
        self.components_x = Xc @ self.rotation_x
        self.components_y = Yc @ self.rotation_y

    # --- API ---
    def reduce_X(self, x_new: np.ndarray) -> np.ndarray:
        if x_new.ndim != 2 or x_new.shape[1] != self.rotation_x.shape[0]:
            raise ValueError(f"x_new must be shape (m, {self.rotation_x.shape[0]}).")
        xc = (x_new - self.mean_x) / self.scale_x
        return xc @ self.rotation_x

    def reduce_Y(self, y_new: np.ndarray) -> np.ndarray:
        if y_new.ndim != 2 or y_new.shape[1] != self.rotation_y.shape[0]:
            raise ValueError(f"y_new must be shape (m, {self.rotation_y.shape[0]}).")
        yc = (y_new - self.mean_y) / self.scale_y
        return yc @ self.rotation_y


from ..Prediction import logistic_regressor
from ..base import *
from ..utils import *
from functools import partial

class logistic_classifier(classification_model):
    def __init__(self, x, y, probability=0.5):
        super().__init__(x, y)
        self.model = logistic_regressor(x, y)
        self.params = self.model.params
        self.probability = probability

    def predict_proba(self, target):
        Fx = self.model.predict(target)
        if isinstance(Fx, (pd.Series, pd.DataFrame)):
            Fx = Fx.values
        Fx = Fx.reshape(-1, 1)
        return np.hstack((1 - Fx, Fx))

    @property
    def tidy(self):
        return self.tidyci(ci=False)

    def tidyci(self, level=0.95, ci=True):
        n = self.n_feat
        if hasattr(self, 'x_vars'):
            df = [self.x_vars, self.model.params[:n], self.model.std_error[:n], self.model.t_statistic[:n],
                  self.model.p_value[:n]]
        else:
            df = [np.arange(n), self.model.params[:n], self.model.std_error[:n], self.model.t_statistic[:n],
                  self.model.p_value[:n]]
        cols = ['variable', 'estimate', 'std.error', 't.statistic', 'p.value']
        if ci:
            df += [self.model.conf_int(level)[:, :n]]
            cols += ['ci.lower', 'ci.upper']
        df = pd.DataFrame(np.vstack(df).T, columns=cols)
        return df

    @property
    def vcov_params(self):
        return self.model.vcov_params

    @property
    def glance(self):
        return pd.DataFrame(self._glance_dict, index=[''])

    @property
    def _glance_dict(self):
        return {'mcfaddens.r.squared': self.model.mcfaddens_r_squared,
                'ben.akiva.lerman.r.squared': self.model.ben_akiva_lerman_r_squared,
                'self.df': self.model.n_feat,
                'resid.df': self.model.degrees_of_freedom,
                'aic': self.model.aic,
                'bic': self.model.bic,
                'log.likelihood': self.model.log_likelihood,
                'deviance': self.model.deviance,
                'probability.threshold': self.probability,
                'accuracy': self.accuracy,
                'misclassification.probability': self.misclassification_probability}


class multinomial_classifier(classification_model):
    def __init__(self, x, y):
        super(multinomial_classifier, self).__init__(x, y)
        self.ohe_y = one_hot_encode(y)
        self.models = [logistic_regressor(x, self.ohe_y[:, i]) for i in range(self.n_classes)]

    def predict_proba(self, target):
        outp = [model.predict(target).reshape(-1, 1) for model in self.models]
        outp = np.hstack(outp)
        outp /= outp.sum(1).reshape(-1, 1)
        return outp

    def tidyci(self, level: float = 0.95, ci: bool = True) -> pd.DataFrame:
        outp = [model.tidyci(level, ci) for model in self.models]
        outp = [pd.concat((pd.DataFrame({'model': np.full(model.shape[0], i)}), model), axis=1) for i, model in
                enumerate(outp)]
        outp = pd.concat(outp, axis=0, ignore_index=True)
        return outp

    @property
    def tidy(self) -> pd.DataFrame: return self.tidyci(ci=False)

    @property
    def glance(self) -> pd.DataFrame:
        outp = [model.glance for model in self.models]
        outp = pd.concat(outp, axis=0)
        outp.index = np.arange(outp.shape[0])
        return outp

LogisticClassifier = partial(PredictionModel,model_type=logistic_classifier)


from ..Distribution import multinomial, two_sample, likelihood_ratio_test
from ..base import *
from ..utils import *


class recursive_partitioning_classifier(classification_model):
    def __init__(self, x, y, sign_level=0.95, max_level=None, random_x=False, level='', classes: Optional[int] = None, weights: Optional[np.ndarray] = None):
        self._col_indx = np.random.permutation(x.shape[1])[
                         :int(np.round(np.sqrt(x.shape[1])))] if random_x else np.arange(x.shape[1])
        x = x[:, self._col_indx]
        super(recursive_partitioning_classifier, self).__init__(x, y)
        if classes != None: self.n_classes = classes
        self.max_level, self.sign_level, self._level, self.weights = max_level, sign_level, level, weights
        self._split_variable, self._split_value, self._p_value, self._left, self._right = np.nan, np.nan, np.nan, None, None
        self._terminal_prediction = multinomial(y, w_x=weights, classes=self.n_classes).params
        if max_level != None and len(level) + 1 == max_level: return
        outp = np.array([self.__get_acc(self.x[:, i]) for i in range(self.n_feat)])
        if np.isnan(outp[:, 0]).all(): return
        self._split_variable = np.nanargmax(outp[:, 0])
        self._split_value = outp[self._split_variable, 1]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._p_value = self._get_p_value()
        if max_level == None and self._p_value > (1 - sign_level) / 2 ** (2 * len(level) + 1): return
        left_indx = x[:, self._split_variable] == 0 if np.isnan(self._split_value) else x[:,
                                                                                        self._split_variable] <= self._split_value
        self._left = recursive_partitioning_classifier(x[left_indx, :], y[left_indx], sign_level=sign_level,
                                                       max_level=max_level, random_x=False, level=level + 'L',
                                                       classes=self.n_classes, weights=weights[left_indx] if weights is not None else None)
        self._right = recursive_partitioning_classifier(x[~left_indx, :], y[~left_indx], sign_level=sign_level,
                                                        max_level=max_level, random_x=False, level=level + 'R',
                                                        classes=self.n_classes, weights=weights[~left_indx] if weights is not None else None)

    # MODIFY WEIGHTS HERE
    def __get_acc(self, var, binary=False):
        if var.var() == 0: return (np.nan, np.nan)
        if binary or np.logical_or(var == 0, var == 1).all():
            retp = [self.y[var == 0], self.y[var == 1]]
            if any([item.shape[0] == 0 for item in retp]): return np.nan, np.nan
            outp = [itemfreq(item).argmax() for item in retp]
            retp = [(retp[i] == outp[i]).sum() for i in range(2)]
            return (sum(retp) / self.n_obs, np.nan)
        splits = np.linspace(var.min(), var.max(), 10)
        acc = np.array([self.__get_acc(var > split, True)[0] for split in splits])
        if np.isnan(acc).all(): return np.full(2, np.nan)
        indx = np.nanargmax(acc)
        return acc[indx], splits[indx]

    def _get_p_value(self):
        xvar = (self.x[:, self._split_variable] == 1) if np.isnan(self._split_value) else (
                self.x[:, self._split_variable] > self._split_value)
        try:
            null = multinomial(self.y, w_x=self.weights)
            if self.weights == None:
                alt = two_sample(x=self.y[xvar], y=self.y[~xvar], model_type=multinomial)
            else:
                alt = two_sample(x=self.y[xvar], y=self.y[~xvar], w_x=self.weights[xvar], w_y=self.weights[~xvar], model_type=multinomial)
            p_value = likelihood_ratio_test(null,alt)
        except:
            return 1
        return p_value['p.value']

    def __str__(self):
        if self._left == None: return f'Level:{self._level} Accuracy:{self.accuracy}; n.obs:{self.n_obs}; Classification:{self._terminal_prediction}; Probability:{self.y.mean()}; p.value:{self._p_value}\n'
        return f'Level:{self._level} n.obs:{self.n_obs}; Variable:{self._col_indx[self._split_variable]}; Split:{self._split_value}; Accuracy:{self.accuracy}; p.value:{self._p_value}\n' + str(
            self._left) + str(self._right)

    def predict_proba(self, newx, fitted=False):
        outp = np.array([self._terminal_prediction] * newx.shape[0])
        if self._left == None: return outp
        if not fitted: newx = newx[:, self._col_indx]
        left_indx = newx[:, self._split_variable] == 0 if np.isnan(self._split_value) else newx[:,
                                                                                           self._split_variable] <= self._split_value
        outp[left_indx,:], outp[~left_indx,:] = self._left.predict_proba(newx[left_indx, :]).reshape(outp[left_indx,:].shape), self._right.predict_proba(newx[~left_indx, :]).reshape(outp[~left_indx,:].shape)
        return outp

    def classify(self, target, fitted=False):
        return self.predict_proba(target, fitted).argmax(1)

    @property
    def fitted(self):
        return self.classify(self.x, fitted=True)



import numpy as np
from ..base import classification_model
from ..utils import itemprob, hstack
from typing import Optional


class naive_bayes(classification_model):
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        priors: Optional[np.ndarray] = None,
        sample_weight: Optional[np.ndarray] = None,
    ) -> None:
        super(naive_bayes, self).__init__(X, y)

        # --- fit priors (optionally weighted) ---
        if priors is None:
            if sample_weight is None:
                self.priors = itemprob(self.y)  # (K,)
            else:
                w = np.asarray(sample_weight, dtype=float).reshape(-1)
                if w.shape[0] != self.n_obs:
                    raise ValueError("sample_weight must have shape (n_obs,).")
                pri = np.array([w[self.y == k].sum() for k in range(self.n_classes)], dtype=float)
                s = pri.sum()
                if s <= 0:
                    raise ValueError("sum of sample weights must be > 0 for prior estimation.")
                self.priors = pri / s
        else:
            self.priors = np.asarray(priors, dtype=float).reshape(-1)
            if self.priors.shape[0] != self.n_classes:
                raise ValueError("priors must have length equal to number of classes.")
            s = self.priors.sum()
            if s <= 0:
                raise ValueError("priors must sum to a positive value.")
            self.priors = self.priors / s

        self._log_priors = np.log(self.priors + 1e-300)  # safety epsilon

    # --- subclass contract ---
    def log_likelihood(self, target: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Subclasses must implement log_likelihood(target).")

    # --- common helpers ---
    def _check_target(self, target: np.ndarray) -> np.ndarray:
        target = np.asarray(target, dtype=float)
        if target.ndim == 1:
            target = target.reshape(1, -1)
        if target.shape[1] != self.n_feat:
            raise ValueError(f"target must have shape (?, {self.n_feat})")
        return target

    def predict_proba(self, target: np.ndarray) -> np.ndarray:
        target = self._check_target(target)
        log_lik = self.log_likelihood(target)  # (n, K)
        log_post = log_lik + self._log_priors  # add log-priors
        m = np.max(log_post, axis=1, keepdims=True)
        stabilized = np.exp(log_post - m)
        probs = stabilized / stabilized.sum(axis=1, keepdims=True)
        return probs

class gaussian_naive_bayes(naive_bayes):
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        priors: Optional[np.ndarray] = None,
        reg: float = 1e-9,
        sample_weight: Optional[np.ndarray] = None,
    ) -> None:
        self.reg = float(reg)
        super(gaussian_naive_bayes, self).__init__(X, y, priors=priors, sample_weight=sample_weight)

        # --- fit per-class means/vars ---
        self.means = np.zeros((self.n_classes, self.n_feat), dtype=float)
        self.vars  = np.zeros((self.n_classes, self.n_feat), dtype=float)

        if sample_weight is None:
            for k in range(self.n_classes):
                Xk = self.x[self.y == k]
                if Xk.shape[0] == 0:
                    continue
                self.means[k, :] = Xk.mean(axis=0)
                self.vars[k,  :] = Xk.var(axis=0) + self.reg
        else:
            w = np.asarray(sample_weight, dtype=float).reshape(-1)
            for k in range(self.n_classes):
                mask = (self.y == k)
                Xk = self.x[mask]
                wk = w[mask]
                if Xk.shape[0] == 0 or wk.sum() == 0:
                    continue
                mk = (wk.reshape(-1, 1) * Xk).sum(axis=0) / wk.sum()
                diff = Xk - mk
                vk = (wk.reshape(-1, 1) * (diff ** 2)).sum(axis=0) / wk.sum()
                self.means[k, :] = mk
                self.vars[k,  :] = vk + self.reg

        # Precompute constants for log-density
        self._log_const = -0.5 * (np.log(2.0 * np.pi) + np.log(self.vars))  # (K, p)

    def log_likelihood(self, target: np.ndarray) -> np.ndarray:
        target = self._check_target(target)
        n = target.shape[0]
        K = self.n_classes
        out = np.empty((n, K), dtype=float)
        for c in range(K):
            diff = target - self.means[c, :]
            term = self._log_const[c, :] - 0.5 * (diff ** 2) / self.vars[c, :]
            out[:, c] = term.sum(axis=1)
        return out

class multinomial_naive_bayes(naive_bayes):
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        priors: Optional[np.ndarray] = None,
        alpha: float = 1.0,
        sample_weight: Optional[np.ndarray] = None,
    ) -> None:
        if np.any(np.asarray(X) < 0):
            raise ValueError("Multinomial NB requires nonnegative features.")
        if alpha < 0:
            raise ValueError("alpha must be nonnegative.")
        self.alpha = float(alpha)

        super(multinomial_naive_bayes, self).__init__(X, y, priors=priors, sample_weight=sample_weight)

        # --- accumulate (possibly weighted) feature counts per class ---
        self.feature_counts = np.zeros((self.n_classes, self.n_feat), dtype=float)

        if sample_weight is None:
            for k in range(self.n_classes):
                Xk = self.x[self.y == k]
                if Xk.size == 0:
                    continue
                self.feature_counts[k, :] = Xk.sum(axis=0)
        else:
            w = np.asarray(sample_weight, dtype=float).reshape(-1)
            if w.shape[0] != self.n_obs:
                raise ValueError("sample_weight must have shape (n_obs,).")
            for k in range(self.n_classes):
                mask = (self.y == k)
                if not np.any(mask):
                    continue
                Xk = self.x[mask]
                wk = w[mask].reshape(-1, 1)
                self.feature_counts[k, :] = (wk * Xk).sum(axis=0)

        # --- Laplace smoothing and normalization per class ---
        smoothed = self.feature_counts + self.alpha
        class_totals = smoothed.sum(axis=1, keepdims=True)  # (K, 1)
        class_totals[class_totals == 0.0] = 1.0  # guard (empty class, alpha==0)
        self.feature_probs = smoothed / class_totals  # (K, p)
        self._log_feature_probs = np.log(self.feature_probs)  # cache

    def log_likelihood(self, target: np.ndarray) -> np.ndarray:
        target = self._check_target(target)
        if np.any(target < 0):
            raise ValueError("Multinomial NB requires nonnegative features.")
        # (n, p) @ (p,) per class -> (n, K)
        n = target.shape[0]
        K = self.n_classes
        out = np.empty((n, K), dtype=float)
        for c in range(K):
            out[:, c] = target @ self._log_feature_probs[c, :].T
        return out


import numpy as np

from .glm import least_squares_regressor, logistic_regressor
from .tree import recursive_partitioning_regressor
from ..utils import *
from ..base import PredictionModel
from functools import partial

class bootstrap_model(ABC):
    def __init__(self, x: np.ndarray, y: np.ndarray, model_type: Type[supervised_model], seed: Optional[int] = None, bootstraps: int = 1000, *args, **kwargs) -> None:
        self.model_type: Type[supervised_model] = model_type
        self.model: supervised_model = model_type(x,y)
        self.seed: Optional[int] = seed
        self.n_boot: int = bootstraps
        self.bootstraps: list[supervised_model] = bootstrap(model_type, x, y, seed=seed, bootstraps=bootstraps)

class bagging_least_squares_regressor(least_squares_regressor):
    def __init__(self, x, y, seed=None, bootstraps=1000):
        super().__init__(x, y)
        self.seed = seed
        self.n_boot = bootstraps
        self.bootstraps = bootstrap(least_squares_regressor, x, y, seed=seed, bootstraps=bootstraps)
        self.bootstrap_params = np.array([item.params for item in self.bootstraps])
        self.params = self.bootstrap_params.mean(0)

    def predict(self, newx):
        return np.array([item.predict(newx) for item in self.bootstraps]).mean(0)\

    @property
    def vcov_params(self):
        return (self.bootstrap_params - self.params).T @ (self.bootstrap_params - self.params) / self.n_boot


class bagging_logistic_regressor(logistic_regressor):
    def __init__(self, x, y, seed=None, bootstraps=1000):
        super().__init__(x, y)
        self.seed = seed
        self.n_boot = bootstraps
        self.bootstraps = bootstrap(logistic_regressor, x, y, seed=seed, bootstraps=bootstraps)
        self.bootstrap_params = np.array([item.params for item in self.bootstraps])
        self.params = self.bootstrap_params.mean(0)

    def predict(self, target: np.ndarray) -> np.ndarray:
        return np.array([item.predict(newx) for item in self.bootstraps]).mean(0)

    @property
    def vcov_params(self):
        return (self.bootstrap_params - self.params).T @ (self.bootstrap_params - self.params) / self.n_boot


class bagging_recursive_partitioning_regressor(recursive_partitioning_regressor):
    def __init__(self, x, y, seed=None, bootstraps=1000, sign_level=0.95, max_level=None, random_x=False):
        super().__init__(x, y, max_level=1)
        self.seed, self.n_boot = seed, bootstraps
        model = lambda x, y: recursive_partitioning_regressor(x, y, sign_level=sign_level, max_level=max_level,
                                                              random_x=random_x)
        self.bootstraps = bootstrap(model, x, y, seed=seed, bootstraps=bootstraps)
        fit = self.fitted
        indx = np.array([np.mean((item.fitted - fit) ** 2) for item in self.bootstraps]).argmin()
        model = self.bootstraps[indx]
        for key, value in vars(model).items(): setattr(self, key, value)

    def predict(self, newx, fitted=False):
        return np.array([item.predict(newx, fitted) for item in self.bootstraps]).mean(0)


random_forest_regressor = partial(bagging_recursive_partitioning_regressor, random_x=True)

BaggingLogisticRegressor = partial(PredictionModel, model_type=bagging_logistic_regressor)
BaggingRecursivePartitioningRegressor = partial(PredictionModel, model_type=bagging_recursive_partitioning_regressor)
RandomForestRegressor = partial(PredictionModel, model_type=random_forest_regressor)
